"""
Example that shows how to use multiple phases in Dymos to model failure of a battery cell
in a simple electrical system.
"""
import unittest

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs
import dymos as dm

from dymos.examples.battery_multibranch.battery_multibranch_ode import BatteryODE


@use_tempdirs
class TestBatteryRKIVP(unittest.TestCase):

    def test_dynamic_params(self):
        prob = om.Problem(model=om.Group())

        traj = prob.model.add_subsystem('traj',  dm.Trajectory())

        # First phase: normal operation.
        # NOTE: using RK4 integration here

        phase0 = dm.Phase(ode_class=BatteryODE, transcription=dm.RungeKutta(num_segments=200))
        phase0.set_time_options(fix_initial=True, fix_duration=True)
        phase0.add_state('state_of_charge', fix_initial=True, fix_final=False,
                         targets=['SOC'], rate_source='dXdt:SOC')
        phase0.add_timeseries_output('battery.V_oc', output_name='V_oc')
        phase0.add_timeseries_output('battery.V_pack', output_name='V_pack')
        phase0.add_timeseries_output('pwr_balance.I_Li', output_name='I_Li')
        traj.add_phase('phase0', phase0)

        # Second phase: normal operation.

        transcription = dm.Radau(num_segments=5, order=5, compressed=True)
        phase1 = dm.Phase(ode_class=BatteryODE, transcription=transcription)
        phase1.set_time_options(fix_initial=False, fix_duration=True)
        phase1.add_state('state_of_charge', fix_initial=False, fix_final=False,
                         solve_segments='forward', targets=['SOC'], rate_source='dXdt:SOC')
        phase1.add_timeseries_output('battery.V_oc', output_name='V_oc', units='V')
        phase1.add_timeseries_output('battery.V_pack', output_name='V_pack', units='V')
        phase1.add_timeseries_output('pwr_balance.I_Li', output_name='I_Li', units='A')
        traj.add_phase('phase1', phase1)

        # Second phase, but with battery failure.

        phase1_bfail = dm.Phase(ode_class=BatteryODE, ode_init_kwargs={'num_battery': 2},
                                transcription=transcription)
        phase1_bfail.set_time_options(fix_initial=False, fix_duration=True)
        phase1_bfail.add_state('state_of_charge', fix_initial=False, fix_final=False,
                               solve_segments='forward', targets=['SOC'], rate_source='dXdt:SOC')
        phase1_bfail.add_timeseries_output('battery.V_oc', output_name='V_oc', units='V')
        phase1_bfail.add_timeseries_output('battery.V_pack', output_name='V_pack', units='V')
        phase1_bfail.add_timeseries_output('pwr_balance.I_Li', output_name='I_Li', units='A')
        traj.add_phase('phase1_bfail', phase1_bfail)

        # Second phase, but with motor failure.

        phase1_mfail = dm.Phase(ode_class=BatteryODE, ode_init_kwargs={'num_motor': 2},
                                transcription=transcription)
        phase1_mfail.set_time_options(fix_initial=False, fix_duration=True)
        phase1_mfail.add_state('state_of_charge', fix_initial=False, fix_final=False,
                               solve_segments='forward', targets=['SOC'], rate_source='dXdt:SOC')
        phase1_mfail.add_timeseries_output('battery.V_oc', output_name='V_oc')
        phase1_mfail.add_timeseries_output('battery.V_pack', output_name='V_pack')
        phase1_mfail.add_timeseries_output('pwr_balance.I_Li', output_name='I_Li')
        traj.add_phase('phase1_mfail', phase1_mfail)

        traj.link_phases(phases=['phase0', 'phase1'], vars=['state_of_charge', 'time'], connected=True)
        traj.link_phases(phases=['phase0', 'phase1_bfail'], vars=['state_of_charge', 'time'], connected=True)
        traj.link_phases(phases=['phase0', 'phase1_mfail'], vars=['state_of_charge', 'time'], connected=True)

        # prob.model.linear_solver = om.DirectSolver(assemble_jac=True)

        prob.setup()
        prob.final_setup()

        prob['traj.phases.phase0.time_extents.t_initial'] = 0
        prob['traj.phases.phase0.time_extents.t_duration'] = 1.0*3600

        # prob['traj.phases.phase1.time_extents.t_initial'] = 1.0*3600
        prob['traj.phases.phase1.time_extents.t_duration'] = 1.0*3600

        # prob['traj.phases.phase1_bfail.time_extents.t_initial'] = 1.0*3600
        prob['traj.phases.phase1_bfail.time_extents.t_duration'] = 1.0*3600

        # prob['traj.phases.phase1_mfail.time_extents.t_initial'] = 1.0*3600
        prob['traj.phases.phase1_mfail.time_extents.t_duration'] = 1.0*3600

        prob.set_solver_print(level=0)
        prob.run_model()

        plot = True
        if plot:
            import matplotlib
            matplotlib.use('Agg')
            import matplotlib.pyplot as plt

            t0 = prob['traj.phase0.timeseries.time']
            t1 = prob['traj.phase1.timeseries.time']
            t1b = prob['traj.phase1_bfail.timeseries.time']
            t1m = prob['traj.phase1_mfail.timeseries.time']
            soc0 = prob['traj.phase0.timeseries.states:state_of_charge']
            soc1 = prob['traj.phase1.timeseries.states:state_of_charge']
            soc1b = prob['traj.phase1_bfail.timeseries.states:state_of_charge']
            soc1m = prob['traj.phase1_mfail.timeseries.states:state_of_charge']

            plt.subplot(2, 2, 1)
            plt.plot(t0, soc0, 'b')
            plt.plot(t1, soc1, 'b')
            plt.plot(t1b, soc1b, 'r')
            plt.plot(t1m, soc1m, 'c')
            plt.xlabel('Time (hour)')
            plt.ylabel('State of Charge (percent)')

            V_oc0 = prob['traj.phase0.timeseries.V_oc']
            V_oc1 = prob['traj.phase1.timeseries.V_oc']
            V_oc1b = prob['traj.phase1_bfail.timeseries.V_oc']
            V_oc1m = prob['traj.phase1_mfail.timeseries.V_oc']

            plt.subplot(2, 2, 2)
            plt.plot(t0, V_oc0, 'b')
            plt.plot(t1, V_oc1, 'b')
            plt.plot(t1b, V_oc1b, 'r')
            plt.plot(t1m, V_oc1m, 'c')
            plt.xlabel('Time (hour)')
            plt.ylabel('Open Circuit Voltage (V)')

            V_pack0 = prob['traj.phase0.timeseries.V_pack']
            V_pack1 = prob['traj.phase1.timeseries.V_pack']
            V_pack1b = prob['traj.phase1_bfail.timeseries.V_pack']
            V_pack1m = prob['traj.phase1_mfail.timeseries.V_pack']

            plt.subplot(2, 2, 3)
            plt.plot(t0, V_pack0, 'b')
            plt.plot(t1, V_pack1, 'b')
            plt.plot(t1b, V_pack1b, 'r')
            plt.plot(t1m, V_pack1m, 'c')
            plt.xlabel('Time (hour)')
            plt.ylabel('Terminal Voltage (V)')

            I_Li0 = prob['traj.phase0.timeseries.I_Li']
            I_Li1 = prob['traj.phase1.timeseries.I_Li']
            I_Li1b = prob['traj.phase1_bfail.timeseries.I_Li']
            I_Li1m = prob['traj.phase1_mfail.timeseries.I_Li']

            plt.subplot(2, 2, 4)
            plt.plot(t0, I_Li0, 'b')
            plt.plot(t1, I_Li1, 'b')
            plt.plot(t1b, I_Li1b, 'r')
            plt.plot(t1m, I_Li1m, 'c')
            plt.xlabel('Time (hour)')
            plt.ylabel('Line Current (A)')

            plt.show()

    def test_static_params(self):
        prob = om.Problem(model=om.Group())

        traj = prob.model.add_subsystem('traj',  dm.Trajectory())

        # First phase: normal operation.
        # NOTE: using RK4 integration here

        P_DEMAND = 2.0

        phase0 = dm.Phase(ode_class=BatteryODE, transcription=dm.RungeKutta(num_segments=200))
        phase0.set_time_options(fix_initial=True, fix_duration=True)
        phase0.add_state('state_of_charge', fix_initial=True, fix_final=False,
                         targets=['SOC'], rate_source='dXdt:SOC')
        phase0.add_timeseries_output('battery.V_oc', output_name='V_oc')
        phase0.add_timeseries_output('battery.V_pack', output_name='V_pack')
        phase0.add_timeseries_output('pwr_balance.I_Li', output_name='I_Li')
        traj.add_phase('phase0', phase0)

        # Second phase: normal operation.

        transcription = dm.Radau(num_segments=5, order=5, compressed=True)
        phase1 = dm.Phase(ode_class=BatteryODE, transcription=transcription)
        phase1.set_time_options(fix_initial=False, fix_duration=True)
        phase1.add_state('state_of_charge', fix_initial=False, fix_final=False,
                         solve_segments='forward', targets=['SOC'], rate_source='dXdt:SOC')
        phase1.add_timeseries_output('battery.V_oc', output_name='V_oc')
        phase1.add_timeseries_output('battery.V_pack', output_name='V_pack')
        phase1.add_timeseries_output('pwr_balance.I_Li', output_name='I_Li')
        traj.add_phase('phase1', phase1)

        # Second phase, but with battery failure.

        phase1_bfail = dm.Phase(ode_class=BatteryODE, ode_init_kwargs={'num_battery': 2},
                                transcription=transcription)
        phase1_bfail.set_time_options(fix_initial=False, fix_duration=True)
        phase1_bfail.add_state('state_of_charge', fix_initial=False, fix_final=False,
                               solve_segments='forward', targets=['SOC'], rate_source='dXdt:SOC')
        phase1_bfail.add_timeseries_output('battery.V_oc', output_name='V_oc')
        phase1_bfail.add_timeseries_output('battery.V_pack', output_name='V_pack')
        phase1_bfail.add_timeseries_output('pwr_balance.I_Li', output_name='I_Li')
        traj.add_phase('phase1_bfail', phase1_bfail)

        # Second phase, but with motor failure.

        phase1_mfail = dm.Phase(ode_class=BatteryODE, ode_init_kwargs={'num_motor': 2},
                                transcription=transcription)
        phase1_mfail.set_time_options(fix_initial=False, fix_duration=True)
        phase1_mfail.add_state('state_of_charge', fix_initial=False, fix_final=False,
                               solve_segments='forward', targets=['SOC'], rate_source='dXdt:SOC')
        phase1_mfail.add_timeseries_output('battery.V_oc', output_name='V_oc')
        phase1_mfail.add_timeseries_output('battery.V_pack', output_name='V_pack')
        phase1_mfail.add_timeseries_output('pwr_balance.I_Li', output_name='I_Li')
        traj.add_phase('phase1_mfail', phase1_mfail)

        traj.link_phases(phases=['phase0', 'phase1'], vars=['state_of_charge', 'time'],
                         connected=True)
        traj.link_phases(phases=['phase0', 'phase1_bfail'], vars=['state_of_charge', 'time'],
                         connected=True)
        traj.link_phases(phases=['phase0', 'phase1_mfail'], vars=['state_of_charge', 'time'],
                         connected=True)

        # prob.model.linear_solver = om.DirectSolver(assemble_jac=True)

        prob.setup()
        prob.final_setup()

        prob['traj.phases.phase0.time_extents.t_initial'] = 0
        prob['traj.phases.phase0.time_extents.t_duration'] = 1.0 * 3600

        # prob['traj.phases.phase1.time_extents.t_initial'] = 1.0*3600
        prob['traj.phases.phase1.time_extents.t_duration'] = 1.0 * 3600

        # prob['traj.phases.phase1_bfail.time_extents.t_initial'] = 1.0*3600
        prob['traj.phases.phase1_bfail.time_extents.t_duration'] = 1.0 * 3600

        # prob['traj.phases.phase1_mfail.time_extents.t_initial'] = 1.0*3600
        prob['traj.phases.phase1_mfail.time_extents.t_duration'] = 1.0 * 3600

        prob.set_solver_print(level=0)
        prob.run_model()

        plot = True
        if plot:
            import matplotlib
            matplotlib.use('Agg')
            import matplotlib.pyplot as plt

            t0 = prob['traj.phase0.timeseries.time']
            t1 = prob['traj.phase1.timeseries.time']
            t1b = prob['traj.phase1_bfail.timeseries.time']
            t1m = prob['traj.phase1_mfail.timeseries.time']
            soc0 = prob['traj.phase0.timeseries.states:state_of_charge']
            soc1 = prob['traj.phase1.timeseries.states:state_of_charge']
            soc1b = prob['traj.phase1_bfail.timeseries.states:state_of_charge']
            soc1m = prob['traj.phase1_mfail.timeseries.states:state_of_charge']

            plt.subplot(2, 2, 1)
            plt.plot(t0, soc0, 'b')
            plt.plot(t1, soc1, 'b')
            plt.plot(t1b, soc1b, 'r')
            plt.plot(t1m, soc1m, 'c')
            plt.xlabel('Time (hour)')
            plt.ylabel('State of Charge (percent)')

            V_oc0 = prob['traj.phase0.timeseries.V_oc']
            V_oc1 = prob['traj.phase1.timeseries.V_oc']
            V_oc1b = prob['traj.phase1_bfail.timeseries.V_oc']
            V_oc1m = prob['traj.phase1_mfail.timeseries.V_oc']

            plt.subplot(2, 2, 2)
            plt.plot(t0, V_oc0, 'b')
            plt.plot(t1, V_oc1, 'b')
            plt.plot(t1b, V_oc1b, 'r')
            plt.plot(t1m, V_oc1m, 'c')
            plt.xlabel('Time (hour)')
            plt.ylabel('Open Circuit Voltage (V)')

            V_pack0 = prob['traj.phase0.timeseries.V_pack']
            V_pack1 = prob['traj.phase1.timeseries.V_pack']
            V_pack1b = prob['traj.phase1_bfail.timeseries.V_pack']
            V_pack1m = prob['traj.phase1_mfail.timeseries.V_pack']

            plt.subplot(2, 2, 3)
            plt.plot(t0, V_pack0, 'b')
            plt.plot(t1, V_pack1, 'b')
            plt.plot(t1b, V_pack1b, 'r')
            plt.plot(t1m, V_pack1m, 'c')
            plt.xlabel('Time (hour)')
            plt.ylabel('Terminal Voltage (V)')

            I_Li0 = prob['traj.phase0.timeseries.I_Li']
            I_Li1 = prob['traj.phase1.timeseries.I_Li']
            I_Li1b = prob['traj.phase1_bfail.timeseries.I_Li']
            I_Li1m = prob['traj.phase1_mfail.timeseries.I_Li']

            plt.subplot(2, 2, 4)
            plt.plot(t0, I_Li0, 'b')
            plt.plot(t1, I_Li1, 'b')
            plt.plot(t1b, I_Li1b, 'r')
            plt.plot(t1m, I_Li1m, 'c')
            plt.xlabel('Time (hour)')
            plt.ylabel('Line Current (A)')

            plt.show()


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np
import openmdao.api as om
import time
from openmdao.utils.array_utils import evenly_distrib_idxs


class vanderpol_ode(om.ExplicitComponent):
    """ODE for optimal control of a Van der Pol oscillator

    objective J:
        minimize integral of (x0**2 + x1**2 + u**2) for 0.0 <= t <= 15

    subject to:
        x0dot = (1 - x1^2) * x0 - x1 + u
        x1dot = x0
        -0.75 <= u <= 1.0

    initial conditions:
        x0(0) = 1.0   x1(0) = 1.0

    final conditions:
        x0(15) = 0.0  x1(15) = 0.0
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # inputs: 2 states and a control
        self.add_input('x0', val=np.ones(nn), desc='derivative of Output', units='V/s')
        self.add_input('x1', val=np.ones(nn), desc='Output', units='V')
        self.add_input('u', val=np.ones(nn), desc='control', units=None)

        # outputs: derivative of states
        # the objective function will be treated as a state for computation, so its derivative is an output
        self.add_output('x0dot', val=np.ones(nn), desc='second derivative of Output', units='V/s**2')
        self.add_output('x1dot', val=np.ones(nn), desc='derivative of Output', units='V/s')
        self.add_output('Jdot', val=np.ones(nn), desc='derivative of objective', units='1.0/s')

        # partials
        r = c = np.arange(nn)

        self.declare_partials(of='x0dot', wrt='x0',  rows=r, cols=c)
        self.declare_partials(of='x0dot', wrt='x1',  rows=r, cols=c)
        self.declare_partials(of='x0dot', wrt='u',   rows=r, cols=c, val=1.0)

        self.declare_partials(of='x1dot', wrt='x0',  rows=r, cols=c, val=1.0)
        self.declare_partials(of='x1dot', wrt='x1',  rows=r, cols=c, val=0.0)
        self.declare_partials(of='x1dot', wrt='u',   rows=r, cols=c, val=0.0)

        self.declare_partials(of='Jdot', wrt='x0',  rows=r, cols=c)
        self.declare_partials(of='Jdot', wrt='x1',  rows=r, cols=c)
        self.declare_partials(of='Jdot', wrt='u',   rows=r, cols=c)

    def compute(self, inputs, outputs):
        x0 = inputs['x0']
        x1 = inputs['x1']
        u = inputs['u']

        outputs['x0dot'] = (1.0 - x1**2) * x0 - x1 + u
        outputs['x1dot'] = x0
        outputs['Jdot'] = x0**2 + x1**2 + u**2

    def compute_partials(self, inputs, jacobian):
        # partials declared with 'val' above do not need to be computed
        x0 = inputs['x0']
        x1 = inputs['x1']
        u = inputs['u']

        jacobian['x0dot', 'x0'] = 1.0 - x1 * x1
        jacobian['x0dot', 'x1'] = -2.0 * x1 * x0 - 1.0

        jacobian['Jdot', 'x0'] = 2.0 * x0
        jacobian['Jdot', 'x1'] = 2.0 * x1
        jacobian['Jdot', 'u'] = 2.0 * u


class vanderpol_ode_group(om.Group):
    """Group containing distributed vanderpol_ode pass through and calculation"""

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_subsystem(name='vanderpol_ode_passthrough',
                           subsys=vanderpol_ode_passthrough(num_nodes=nn),
                           promotes_inputs=['x0', 'x1', 'u'])

        self.add_subsystem(name='vanderpol_ode_delay',
                           subsys=vanderpol_ode_delay(num_nodes=nn),
                           promotes_outputs=['x0dot', 'x1dot', 'Jdot'])

        # connect collect_comp (pass through) output to distributed ODE input
        self.connect('vanderpol_ode_passthrough.x0pass', 'vanderpol_ode_delay.x0')
        self.connect('vanderpol_ode_passthrough.x1pass', 'vanderpol_ode_delay.x1')
        self.connect('vanderpol_ode_passthrough.upass', 'vanderpol_ode_delay.u')


class vanderpol_ode_passthrough(om.ExplicitComponent):
    """Pass through component that just copies control and state from input to output

    if you just use a plain old passthrough (non-distributed with a full sized input and output) component to connect
    to the distributed output, the framework will do the MPI allgathering for you"""

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']  # total number of inputs and outputs over all processes

        # inputs: 2 states and a control
        self.add_input('x0', val=np.ones(nn), desc='derivative of Output', units='V/s')
        self.add_input('x1', val=np.ones(nn), desc='Output', units='V')
        self.add_input('u', val=np.ones(nn), desc='control', units=None)

        # outputs: same as inputs
        self.add_output('x0pass', val=np.ones(nn), desc='derivative of Output', units='V/s')
        self.add_output('x1pass', val=np.ones(nn), desc='Output', units='V')
        self.add_output('upass', val=np.ones(nn), desc='control', units=None)

        # partials
        row_col = np.arange(nn)
        self.declare_partials(of='x0pass', wrt='x0', rows=row_col, cols=row_col, val=1.0)
        self.declare_partials(of='x1pass', wrt='x1', rows=row_col, cols=row_col, val=1.0)
        self.declare_partials(of='upass', wrt='u', rows=row_col, cols=row_col, val=1.0)

    def compute(self, inputs, outputs):
        outputs['x0pass'] = inputs['x0']
        outputs['x1pass'] = inputs['x1']
        outputs['upass'] = inputs['u']


class vanderpol_ode_delay(om.ExplicitComponent):
    """intentionally slow version of vanderpol_ode for effects of demonstrating distributed component calculations

    MPI can run this component in multiple processes, distributing the calculation of derivatives.
    This code has a delay in it to simulate a longer computation. It should run faster with more processes.
    """

    def __init__(self, *args, **kwargs):
        self.delay_time = 0.005
        self.progress_prints = False
        super().__init__(*args, **kwargs)

    def initialize(self):
        self.options.declare('num_nodes', types=int)
        self.options['distributed'] = True
        self.options.declare('size', types=int, default=1, desc="Size of input and output vectors.")

    def setup(self):
        nn = self.options['num_nodes']
        comm = self.comm
        rank = comm.rank

        sizes, offsets = evenly_distrib_idxs(comm.size, nn)  # (#cpus, #inputs) -> (size array, offset array)
        start = offsets[rank]
        self.io_size = sizes[rank]  # number of inputs and outputs managed by this distributed process
        end = start + self.io_size

        if self.progress_prints:
            print('in vanderpol_ode_delay.setup', self.io_size, self.comm.rank)

        # inputs: 2 states and a control
        self.add_input('x0', val=np.ones(self.io_size), desc='derivative of Output', units='V/s',
                       src_indices=np.arange(start, end, dtype=int))
        self.add_input('x1', val=np.ones(self.io_size), desc='Output', units='V',
                       src_indices=np.arange(start, end, dtype=int))
        self.add_input('u', val=np.ones(self.io_size), desc='control', units=None,
                       src_indices=np.arange(start, end, dtype=int))

        # outputs: derivative of states
        # the objective function will be treated as a state for computation, so its derivative is an output
        self.add_output('x0dot', val=np.ones(self.io_size), desc='second derivative of Output', units='V/s**2')
        self.add_output('x1dot', val=np.ones(self.io_size), desc='derivative of Output', units='V/s')
        self.add_output('Jdot', val=np.ones(self.io_size), desc='derivative of objective', units='1.0/s')

        # partials
        r = c = np.arange(self.io_size)

        self.declare_partials(of='x0dot', wrt='x0',  rows=r, cols=c)
        self.declare_partials(of='x0dot', wrt='x1',  rows=r, cols=c)
        self.declare_partials(of='x0dot', wrt='u',   rows=r, cols=c, val=1.0)

        self.declare_partials(of='x1dot', wrt='x0',  rows=r, cols=c, val=1.0)
        self.declare_partials(of='x1dot', wrt='x1',  rows=r, cols=c, val=0.0)
        self.declare_partials(of='x1dot', wrt='u',   rows=r, cols=c, val=0.0)

        self.declare_partials(of='Jdot', wrt='x0',  rows=r, cols=c)
        self.declare_partials(of='Jdot', wrt='x1',  rows=r, cols=c)
        self.declare_partials(of='Jdot', wrt='u',   rows=r, cols=c)

    def compute(self, inputs, outputs):
        if self.progress_prints:
            sizes = (len(inputs['x0']), len(inputs['x1']), len(inputs['u']))
            print('in vanderpol_ode_delay.compute', sizes, self.comm.rank)

        time.sleep(self.delay_time * self.io_size)  # introduce slowness proportional to size of computation

        x0 = inputs['x0']
        x1 = inputs['x1']
        u = inputs['u']

        outputs['x0dot'] = (1.0 - x1**2) * x0 - x1 + u
        outputs['x1dot'] = x0
        outputs['Jdot'] = x0**2 + x1**2 + u**2

    def compute_partials(self, inputs, jacobian):
        if self.progress_prints:
            sizes = (len(inputs['x0']), len(inputs['x1']), len(inputs['u']))
            print('in vanderpol_ode_delay.compute_partials', sizes)

        time.sleep(self.delay_time * self.io_size)  # introduce slowness proportional to size of computation

        # partials declared with 'val' above do not need to be computed
        x0 = inputs['x0']
        x1 = inputs['x1']
        u = inputs['u']

        jacobian['x0dot', 'x0'] = 1.0 - x1 * x1
        jacobian['x0dot', 'x1'] = -2.0 * x1 * x0 - 1.0

        jacobian['Jdot', 'x0'] = 2.0 * x0
        jacobian['Jdot', 'x1'] = 2.0 * x1
        jacobian['Jdot', 'u'] = 2.0 * u



import openmdao.api as om
import dymos as dm
from dymos.examples.vanderpol.vanderpol_ode import vanderpol_ode, vanderpol_ode_group


def vanderpol(transcription='gauss-lobatto', num_segments=8, transcription_order=3,
              compressed=True, optimizer='SLSQP', use_pyoptsparse=False, delay=None):
    """Dymos problem definition for optimal control of a Van der Pol oscillator"""

    # define the OpenMDAO problem
    p = om.Problem(model=om.Group())

    if not use_pyoptsparse:
        p.driver = om.ScipyOptimizeDriver()
    else:
        p.driver = om.pyOptSparseDriver()
    p.driver.options['optimizer'] = optimizer
    if use_pyoptsparse:
        if optimizer == 'SNOPT':
            p.driver.opt_settings['iSumm'] = 6  # show detailed SNOPT output
        elif optimizer == 'IPOPT':
            p.driver.opt_settings['print_level'] = 4
    p.driver.declare_coloring()

    # define a Trajectory object and add to model
    traj = dm.Trajectory()
    p.model.add_subsystem('traj', subsys=traj)

    # define a Transcription
    if transcription == 'gauss-lobatto':
        t = dm.GaussLobatto(num_segments=num_segments,
                            order=transcription_order,
                            compressed=compressed)
    elif transcription == 'radau-ps':
        t = dm.Radau(num_segments=num_segments,
                     order=transcription_order,
                     compressed=compressed)
    elif transcription == 'runge-kutta':
        t = dm.RungeKutta(num_segments=num_segments,
                          order=transcription_order,
                          compressed=compressed)

    # define a Phase as specified above and add to Phase
    if not delay:
        phase = dm.Phase(ode_class=vanderpol_ode, transcription=t)
    else:
        phase = dm.Phase(ode_class=vanderpol_ode_group, transcription=t)  # distributed component group
    traj.add_phase(name='phase0', phase=phase)

    t_final = 15.0
    phase.set_time_options(fix_initial=True, fix_duration=True, duration_val=t_final, units='s')

    # set the State time options
    phase.add_state('x0', fix_initial=False, fix_final=False,
                    rate_source='x0dot',
                    units='V/s',
                    targets='x0')  # target required because x0 is an input
    phase.add_state('x1', fix_initial=False, fix_final=False,
                    rate_source='x1dot',
                    units='V',
                    targets='x1')  # target required because x1 is an input
    phase.add_state('J', fix_initial=False, fix_final=False,
                    rate_source='Jdot',
                    units=None)

    # define the control
    phase.add_control(name='u', units=None, lower=-0.75, upper=1.0, continuity=True,
                      rate_continuity=True, targets='u')  # target required because u is an input

    # add constraints
    phase.add_boundary_constraint('x0', loc='initial', equals=1.0)
    phase.add_boundary_constraint('x1', loc='initial', equals=1.0)
    phase.add_boundary_constraint('J', loc='initial', equals=0.0)

    phase.add_boundary_constraint('x0', loc='final', equals=0.0)
    phase.add_boundary_constraint('x1', loc='final', equals=0.0)

    # define objective to minimize
    phase.add_objective('J', loc='final')

    # setup the problem
    p.setup(check=True)

    # TODO - Dymos API will soon provide a way to specify this.
    # the linear solver used to compute derivatives is not working on MPI, so switch to LinearRunOnce
    for phase in traj._phases.values():
        phase.linear_solver = om.LinearRunOnce()

    p['traj.phase0.t_initial'] = 0.0
    p['traj.phase0.t_duration'] = t_final

    # add a linearly interpolated initial guess for the state and control curves
    p['traj.phase0.states:x0'] = phase.interpolate(ys=[1, 0], nodes='state_input')
    p['traj.phase0.states:x1'] = phase.interpolate(ys=[1, 0], nodes='state_input')
    p['traj.phase0.states:J'] = phase.interpolate(ys=[0, 1], nodes='state_input')
    p['traj.phase0.controls:u'] = phase.interpolate(ys=[-0.75, -0.75], nodes='control_input')

    p.final_setup()

    # debugging helpers:
    # om.n2(p)       # show n2 diagram
    #
    # with np.printoptions(linewidth=1024):  # display partials for manual checking
    #     p.check_partials(compact_print=True)

    return p


if __name__ == '__main__':
    # just set up the problem, test it elsewhere
    p = vanderpol(transcription='gauss-lobatto', num_segments=75, transcription_order=3,
                  compressed=True, optimizer='SLSQP')



import warnings

import numpy as np

import openmdao.api as om
from openmdao.utils.general_utils import warn_deprecation

from ..transcription_base import TranscriptionBase
from .components import RungeKuttaStepsizeComp, RungeKuttaStateContinuityIterGroup, \
    RungeKuttaTimeseriesOutputComp, RungeKuttaControlContinuityComp
from ..common import TimeComp, PathConstraintComp
from ...utils.rk_methods import rk_methods
from ...utils.misc import CoerceDesvar, get_rate_units, get_source_metadata
from ...utils.introspection import get_targets
from ...utils.constants import INF_BOUND
from ...utils.indexing import get_src_indices_by_row
from ..grid_data import GridData
from fnmatch import filter


class RungeKutta(TranscriptionBase):
    """
    The RungeKutta Transcription class.

    RungeKutta transcription in Dymos uses the RungeKutta-based shooting method which propagates
    the states from the phase initial time to the phase final time.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super(RungeKutta, self).__init__(**kwargs)
        self._rhs_source = 'ode'

        msg = 'The RungeKutta transcription is deprecated and will be removed in Dymos v1.0.0.\n' \
              'For equivalent behavior, users should switch to ' \
              'GaussLobatto(order=3, solve_segments=\'forward\') or ' \
              'GaussLobatto(order=3, solve_segments=\'backward\')'
        warn_deprecation(msg)

    def initialize(self):
        """
        Declare transcription options.
        """
        self.options.declare('method', default='RK4', values=('RK4',),
                             desc='The integrator used within the explicit phase.')

        self.options.declare('k_solver_class', default=om.NonlinearBlockGS,
                             values=(om.NonlinearBlockGS, om.NewtonSolver, om.NonlinearRunOnce),
                             allow_none=True,
                             desc='The nonlinear solver class used to converge the numerical '
                                  'integration across each segment.')

        self.options.declare('k_solver_options', default={'iprint': -1}, types=(dict,),
                             desc='The options passed to the nonlinear solver used to converge the'
                                  'Runge-Kutta propagation across each step.')

    def init_grid(self):
        """
        Setup the GridData object for the Transcription.
        """
        self.grid_data = GridData(num_segments=self.options['num_segments'],
                                  transcription='runge-kutta',
                                  transcription_order=self.options['method'],
                                  segment_ends=self.options['segment_ends'],
                                  compressed=self.options['compressed'])

    def setup_solvers(self, phase):
        """
        Add a NewtonSolver to converge continuity errors in the state between steps.

        Parameters
        ----------
        phase : dymos.Phase
            The phase to which this transcription instance applies.
        """
        phase.nonlinear_solver = om.NewtonSolver()
        phase.nonlinear_solver.options['iprint'] = -1
        phase.nonlinear_solver.options['solve_subsystems'] = True
        phase.nonlinear_solver.options['err_on_non_converge'] = True
        phase.nonlinear_solver.linesearch = om.BoundsEnforceLS()

    def configure_solvers(self, phase):
        """
        Configure the solvers.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def setup_time(self, phase):
        """
        Setup the time component.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        time_units = phase.time_options['units']
        num_seg = self.options['num_segments']
        rk_data = rk_methods[self.options['method']]
        num_nodes = num_seg * rk_data['num_stages']
        grid_data = self.grid_data

        super(RungeKutta, self).setup_time(phase)

        time_comp = TimeComp(num_nodes=num_nodes, node_ptau=grid_data.node_ptau,
                             node_dptau_dstau=grid_data.node_dptau_dstau, units=time_units)

        phase.add_subsystem('time', time_comp, promotes_outputs=['*'], promotes_inputs=['*'])

        h_comp = RungeKuttaStepsizeComp(num_segments=num_seg,
                                        seg_rel_lengths=np.diff(grid_data.segment_ends),
                                        time_units=time_units)

        phase.add_subsystem('stepsize_comp',
                            subsys=h_comp,
                            promotes_inputs=['t_duration'],
                            promotes_outputs=['h'])

    def configure_time(self, phase):
        """
        Configure the inputs/outputs on the time component.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(RungeKutta, self).configure_time(phase)

        phase.time.configure_io()
        phase.stepsize_comp.configure_io()

        options = phase.time_options

        # The tuples here are (name, user_specified_targets, dynamic)
        for name, usr_tgts, dynamic in [('time', options['targets'], True),
                                        ('time_phase', options['time_phase_targets'], True),
                                        ('t_initial', options['t_initial_targets'], False),
                                        ('t_duration', options['t_duration_targets'], False)]:

            targets = get_targets(phase.ode, name=name, user_targets=usr_tgts)
            if targets:
                all_src_idxs = self.grid_data.subset_node_indices['all'] if dynamic else None
                end_src_idxs = self.grid_data.subset_node_indices['segment_ends'] if dynamic else None
                phase.connect(name, ['rk_solve_group.ode.{0}'.format(t) for t in targets],
                              src_indices=all_src_idxs)
                phase.connect(name, ['ode.{0}'.format(t) for t in targets],
                              src_indices=end_src_idxs)

    def setup_ode(self, phase):
        """
        Setup the ode for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        phase.add_subsystem('rk_solve_group',
                            RungeKuttaStateContinuityIterGroup(
                                num_segments=self.options['num_segments'],
                                method=self.options['method'],
                                state_options=phase.state_options,
                                time_units=phase.time_options['units'],
                                ode_class=phase.options['ode_class'],
                                ode_init_kwargs=phase.options['ode_init_kwargs'],
                                k_solver_class=self.options['k_solver_class'],
                                k_solver_options=self.options['k_solver_options']))

        # Since the RK Solve group evaluates the ODE at *predicted* state values, we need
        # to instantiate a second ODE group that will call the ODE at the actual integrated
        # state values so that we can accurate evaluate path and boundary constraints and
        # obtain timeseries for ODE outputs.
        phase.add_subsystem('ode',
                            phase.options['ode_class'](num_nodes=2*self.options['num_segments'],
                                                       **phase.options['ode_init_kwargs']))

    def configure_ode(self, phase):
        """
        Create connections to the introspected states.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        phase.rk_solve_group.configure_io()

        num_connected = len([s for s in phase.state_options
                             if phase.state_options[s]['connected_initial']])

        phase.promotes('rk_solve_group',
                       inputs=['h'] if num_connected == 0 else ['h', 'initial_states:*'],
                       outputs=['states:*', 'state_predict_comp.predicted_states:*'])

    def _get_rate_source_path(self, state_name, phase, nodes=None, **kwargs):
        try:
            var = phase.state_options[state_name]['rate_source']
        except RuntimeError:
            raise ValueError('state \'{0}\' in phase \'{1}\' was not given a '
                             'rate_source'.format(state_name, phase.name))
        shape = phase.state_options[state_name]['shape']
        state_size = np.prod(shape)
        var_type = phase.classify_var(var)
        num_segments = self.options['num_segments']
        num_stages = rk_methods[self.options['method']]['num_stages']

        # Determine the path to the variable
        if var_type == 'time':
            rate_path = 'time'
            src_idxs = None
        elif var_type == 'time_phase':
            rate_path = 'time_phase'
            src_idxs = None
        elif var_type == 'state':
            rate_path = 'state_predict_comp.predicted_states:{0}'.format(var)
            size = num_segments * num_stages * state_size
            src_idxs = np.arange(size, dtype=int).reshape((num_segments, num_stages, state_size))
        elif var_type == 'indep_control':
            rate_path = 'control_values:{0}'.format(var)
            size = num_segments * num_stages * state_size
            src_idxs = np.arange(size, dtype=int).reshape((num_segments, num_stages, state_size))
        elif var_type == 'input_control':
            rate_path = 'control_values:{0}'.format(var)
            size = num_segments * num_stages * state_size
            src_idxs = np.arange(size, dtype=int).reshape((num_segments, num_stages, state_size))
        elif var_type == 'control_rate':
            rate_path = 'control_rates:{0}'.format(var)
            size = num_segments * num_stages * state_size
            src_idxs = np.arange(size, dtype=int).reshape((num_segments, num_stages, state_size))
        elif var_type == 'control_rate2':
            rate_path = 'control_rates:{0}'.format(var)
            size = num_segments * num_stages * state_size
            src_idxs = np.arange(size, dtype=int).reshape((num_segments, num_stages, state_size))
        elif var_type == 'indep_polynomial_control':
            rate_path = 'polynomial_control_values:{0}'.format(var)
            size = num_segments * num_stages * state_size
            src_idxs = np.arange(size, dtype=int).reshape((num_segments, num_stages, state_size))
        elif var_type == 'input_polynomial_control':
            rate_path = 'polynomial_control_values:{0}'.format(var)
            size = num_segments * num_stages * state_size
            src_idxs = np.arange(size, dtype=int).reshape((num_segments, num_stages, state_size))
        elif var_type == 'polynomial_control_rate':
            control_name = var[:-5]
            rate_path = 'polynomial_control_rates:{0}_rate'.format(control_name)
            size = num_segments * num_stages * state_size
            src_idxs = np.arange(size, dtype=int).reshape((num_segments, num_stages, state_size))
        elif var_type == 'polynomial_control_rate2':
            control_name = var[:-6]
            rate_path = 'polynomial_control_rates:{0}_rate2'.format(control_name)
            size = num_segments * num_stages * state_size
            src_idxs = np.arange(size, dtype=int).reshape((num_segments, num_stages, state_size))
        elif var_type == 'parameter':
            rate_path = 'parameters:{0}'.format(var)
            size = np.prod(phase.parameter_options[var]['shape'])
            src_idxs = np.zeros(num_segments * num_stages * size, dtype=int).reshape((num_segments,
                                                                                      num_stages,
                                                                                      state_size))
        else:
            # Failed to find variable, assume it is in the ODE
            rate_path = 'rk_solve_group.ode.{0}'.format(var)
            state_size = np.prod(shape)
            size = num_segments * num_stages * state_size
            src_idxs = np.arange(size, dtype=int).reshape((num_segments, num_stages, state_size))

        return rate_path, src_idxs

    def setup_states(self, phase):
        """
        Setup the states for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_states(self, phase):
        """
        Configure state connections post-introspection.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        num_seg = self.options['num_segments']
        num_state_input_nodes = num_seg + 1

        for state_name, options in phase.state_options.items():

            self._configure_state_introspection(state_name, options, phase)

            size = np.prod(options['shape'])

            if options['opt']:
                # Set the desvar indices accordingly
                desvar_indices = list(range(size))

                if options['fix_initial']:
                    if options['initial_bounds'] is not None:
                        raise ValueError('Cannot specify \'fix_initial=True\' and specify '
                                         'initial_bounds for state {0}'.format(state_name))
                    if options['connected_initial']:
                        raise ValueError('Cannot specify \'fix_initial=True\' and specify '
                                         '\'connected_initial=True\' for state {0} in '
                                         'phase {1}.'.format(state_name, phase.name))
                    del desvar_indices[:size]

                if options['fix_final']:
                    raise ValueError('Cannot specify \'fix_final=True\' in '
                                     'RungeKuttaPhase'.format(state_name))

                if options['final_bounds'] is not None:
                    raise ValueError('Cannot specify \'final_bounds\' in RungeKuttaPhase '
                                     '(state {0})'.format(state_name))

                if len(desvar_indices) > 0:
                    coerce_desvar_option = CoerceDesvar(num_state_input_nodes, desvar_indices,
                                                        options)

                    lb = np.zeros_like(desvar_indices, dtype=float)
                    lb[:] = -INF_BOUND if coerce_desvar_option('lower') is None else \
                        coerce_desvar_option('lower')

                    ub = np.zeros_like(desvar_indices, dtype=float)
                    ub[:] = INF_BOUND if coerce_desvar_option('upper') is None else \
                        coerce_desvar_option('upper')

                    if options['initial_bounds'] is not None:
                        lb[0] = options['initial_bounds'][0]
                        ub[0] = options['initial_bounds'][-1]

                    phase.add_design_var(name='states:{0}'.format(state_name),
                                         lower=lb,
                                         upper=ub,
                                         scaler=coerce_desvar_option('scaler'),
                                         adder=coerce_desvar_option('adder'),
                                         ref0=coerce_desvar_option('ref0'),
                                         ref=coerce_desvar_option('ref'),
                                         indices=desvar_indices)

        num_seg = self.options['num_segments']

        for state_name, options in phase.state_options.items():
            shape = options['shape']

            # Connect the states at the segment ends to the final ODE instance.
            row_idxs = np.repeat(np.arange(1, num_seg, dtype=int), repeats=2)
            row_idxs = np.concatenate(([0], row_idxs, [num_seg]))
            src_idxs = get_src_indices_by_row(row_idxs, shape)

            if shape == (1,):
                src_idxs = src_idxs.ravel()

            targets = get_targets(ode=phase.ode, name=state_name, user_targets=options['targets'])

            if targets:
                phase.connect('states:{0}'.format(state_name),
                              ['ode.{0}'.format(tgt) for tgt in targets],
                              src_indices=src_idxs, flat_src_indices=True)

            # Connect the state rate source to the k comp
            rate_path, src_idxs = self._get_rate_source_path(state_name, phase)

            phase.connect(rate_path,
                          'rk_solve_group.k_comp.f:{0}'.format(state_name),
                          src_indices=src_idxs,
                          flat_src_indices=True)

    def setup_controls(self, phase):
        """
        Setup the control group.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(RungeKutta, self).setup_controls(phase)

    def configure_controls(self, phase):
        """
        Configure the inputs/outputs for the controls.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(RungeKutta, self).configure_controls(phase)

        grid_data = self.grid_data

        for name, options in phase.control_options.items():
            shape = options['shape']
            segment_end_idxs = grid_data.subset_node_indices['segment_ends']
            all_idxs = grid_data.subset_node_indices['all']
            segend_src_idxs = get_src_indices_by_row(segment_end_idxs, shape=shape)
            all_src_idxs = get_src_indices_by_row(all_idxs, shape=shape)

            if shape == (1,):
                segend_src_idxs = segend_src_idxs.ravel()
                all_src_idxs = all_src_idxs.ravel()

            targets = get_targets(ode=phase.ode, name=name, user_targets=options['targets'])
            if targets:
                src_name = 'control_values:{0}'.format(name)
                phase.connect(src_name,
                              ['ode.{0}'.format(t) for t in targets],
                              src_indices=segend_src_idxs, flat_src_indices=True)
                phase.connect(src_name,
                              ['rk_solve_group.ode.{0}'.format(t) for t in targets],
                              src_indices=all_src_idxs, flat_src_indices=True)

            targets = get_targets(ode=phase.ode, name=f'{name}_rate',
                                  user_targets=options['rate_targets'])
            if targets:
                src_name = 'control_rates:{0}_rate'.format(name)

                phase.connect(src_name,
                              ['ode.{0}'.format(t) for t in targets],
                              src_indices=segend_src_idxs, flat_src_indices=True)

                phase.connect(src_name,
                              ['rk_solve_group.ode.{0}'.format(t) for t in targets],
                              src_indices=all_src_idxs, flat_src_indices=True)

            targets = get_targets(ode=phase.ode, name=f'{name}_rate2',
                                  user_targets=options['rate2_targets'])
            if targets:
                src_name = 'control_rates:{0}_rate2'.format(name)

                phase.connect(src_name,
                              ['ode.{0}'.format(t) for t in targets],
                              src_indices=segend_src_idxs, flat_src_indices=True)

                phase.connect(src_name,
                              ['rk_solve_group.ode.{0}'.format(t) for t in targets],
                              src_indices=all_src_idxs, flat_src_indices=True)

    def setup_polynomial_controls(self, phase):
        """
        Adds the polynomial control group to the model if any polynomial controls are present.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(RungeKutta, self).setup_polynomial_controls(phase)

    def configure_polynomial_controls(self, phase):
        """
        Configure the inputs/outputs for the polynomial controls.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(RungeKutta, self).configure_polynomial_controls(phase)
        grid_data = self.grid_data

        for name, options in phase.polynomial_control_options.items():
            shape = options['shape']
            segment_end_idxs = grid_data.subset_node_indices['segment_ends']
            all_idxs = grid_data.subset_node_indices['all']
            segend_src_idxs = get_src_indices_by_row(segment_end_idxs, shape=shape)
            all_src_idxs = get_src_indices_by_row(all_idxs, shape=shape)

            if shape == (1,):
                segend_src_idxs = segend_src_idxs.ravel()
                all_src_idxs = all_src_idxs.ravel()

            targets = get_targets(phase.ode, name=name, user_targets=options['targets'])
            if targets:
                src_name = 'polynomial_control_values:{0}'.format(name)
                phase.connect(src_name,
                              ['ode.{0}'.format(t) for t in targets],
                              src_indices=segend_src_idxs, flat_src_indices=True)

                phase.connect(src_name,
                              ['rk_solve_group.ode.{0}'.format(t) for t in targets],
                              src_indices=all_src_idxs, flat_src_indices=True)

            targets = get_targets(ode=phase.ode, name=f'{name}_rate',
                                  user_targets=options['rate_targets'])
            if targets:
                phase.connect(f'polynomial_control_rates:{name}_rate',
                              ['ode.{0}'.format(t) for t in targets],
                              src_indices=segend_src_idxs, flat_src_indices=True)
                phase.connect(f'polynomial_control_rates:{name}_rate',
                              ['rk_solve_group.ode.{0}'.format(t) for t in targets],
                              src_indices=all_src_idxs, flat_src_indices=True)

            targets = get_targets(ode=phase.ode, name=f'{name}_rate2',
                                  user_targets=options['rate2_targets'])
            if targets:
                phase.connect(f'polynomial_control_rates:{name}_rate2',
                              [f'ode.{t}' for t in targets],
                              src_indices=segend_src_idxs, flat_src_indices=True)
                phase.connect(f'polynomial_control_rates:{name}_rate2',
                              [f'rk_solve_group.ode.{t}' for t in targets],
                              src_indices=all_src_idxs, flat_src_indices=True)

    def setup_defects(self, phase):
        """
        Setup the Collocation and Continuity components as necessary.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        grid_data = self.grid_data
        num_seg = grid_data.num_segments

        # Add the continuity constraint component if necessary
        if num_seg > 1 and phase.control_options:
            time_units = phase.time_options['units']

            phase.add_subsystem('continuity_comp',
                                RungeKuttaControlContinuityComp(grid_data=grid_data,
                                                                state_options=phase.state_options,
                                                                control_options=phase.control_options,
                                                                time_units=time_units),
                                promotes_inputs=['t_duration'])

    def configure_defects(self, phase):
        """
        Configure the continuity_comp and connect the collocation constraints.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        grid_data = self.grid_data
        num_seg = grid_data.num_segments

        # Add the continuity constraint component if necessary
        if num_seg > 1 and phase.control_options:
            phase.continuity_comp.configure_io()

            for name, options in phase.control_options.items():
                # The sub-indices of control_disc indices that are segment ends
                segment_end_idxs = grid_data.subset_node_indices['segment_ends']
                src_idxs = get_src_indices_by_row(segment_end_idxs, options['shape'], flat=True)

                phase.connect('control_values:{0}'.format(name),
                              'continuity_comp.controls:{0}'.format(name),
                              src_indices=src_idxs, flat_src_indices=True)

                phase.connect('control_rates:{0}_rate'.format(name),
                              'continuity_comp.control_rates:{}_rate'.format(name),
                              src_indices=src_idxs, flat_src_indices=True)

                phase.connect('control_rates:{0}_rate2'.format(name),
                              'continuity_comp.control_rates:{}_rate2'.format(name),
                              src_indices=src_idxs, flat_src_indices=True)

    def setup_path_constraints(self, phase):
        """
        Add a path constraint component if necessary and issue appropriate connections.

        This overrides the default transcription path constraints at
        all nodes and only applies them at the segment end points, the only points in an RK segment
        where all of the variables are valid.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        gd = self.grid_data

        if phase._path_constraints:
            path_comp = PathConstraintComp(num_nodes=gd.subset_num_nodes['segment_ends'])
            phase.add_subsystem('path_constraints', subsys=path_comp)

    def configure_path_constraints(self, phase):
        """
        Handle the common operations for configuration of the path constraints.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(RungeKutta, self).configure_path_constraints(phase)

        gd = self.grid_data
        num_seg = gd.num_segments
        num_seg_ends = self.grid_data.subset_num_nodes['segment_ends']
        seg_end_idxs = self.grid_data.subset_node_indices['segment_ends']

        for var, options in phase._path_constraints.items():
            con_name = options['constraint_name']

            # Determine the path to the variable which we will be constraining
            # This is more complicated for path constraints since, for instance,
            # a single state variable has two sources which must be connected to
            # the path component.
            var_type = phase.classify_var(var)

            if var_type == 'time':
                src = 'time'
                tgt = f'path_constraints.all_values:{con_name}'
                src_idxs = np.reshape(seg_end_idxs, newshape=(num_seg_ends, 1))
                flat_src_idxs = False

            elif var_type == 'time_phase':
                src = 'time_phase'
                tgt = f'path_constraints.all_values:{con_name}'
                src_idxs = np.reshape(seg_end_idxs, newshape=(num_seg_ends, 1))
                flat_src_idxs = False

            elif var_type == 'state':
                state_shape = phase.state_options[var]['shape']
                row_idxs = np.repeat(np.arange(1, num_seg, dtype=int), repeats=2)
                row_idxs = np.concatenate(([0], row_idxs, [num_seg]))
                src_idxs = get_src_indices_by_row(row_idxs, state_shape)
                flat_src_idxs = True
                src = f'states:{var}'
                tgt = f'path_constraints.all_values:{var}'

            elif var_type in ('indep_control', 'input_control'):
                control_shape = phase.control_options[var]['shape']

                src_rows = self.grid_data.subset_node_indices['segment_ends']
                src_idxs = get_src_indices_by_row(src_rows, shape=control_shape)
                flat_src_idxs = True

                src = f'control_values:{var}'
                tgt = f'path_constraints.all_values:{con_name}'

            elif var_type in ('indep_polynomial_control', 'input_polynomial_control'):
                shape = phase.polynomial_control_options[var]['shape']
                src_idxs = get_src_indices_by_row(seg_end_idxs, shape=shape)
                flat_src_idxs = True

                src = f'polynomial_control_values:{var}'
                tgt = f'path_constraints.all_values:{con_name}'

            elif var_type in ('control_rate', 'control_rate2'):
                control_name = var[:-5] if var.endswith('_rate') else var[:-6]
                control_shape = phase.control_options[control_name]['shape']
                src_idxs = get_src_indices_by_row(seg_end_idxs, shape=control_shape)
                flat_src_idxs = True

                src = f'control_rates:{var}'
                tgt = f'path_constraints.all_values:{con_name}'

            elif var_type in ('polynomial_control_rate', 'polynomial_control_rate2'):
                control_name = var[:-5] if var.endswith('_rate') else var[:-6]
                control_shape = phase.polynomial_control_options[control_name]['shape']
                src_idxs = get_src_indices_by_row(seg_end_idxs, shape=control_shape)
                flat_src_idxs = True

                src = f'polynomial_control_rates:{var}'
                tgt = f'path_constraints.all_values:{con_name}'

            else:
                src_rows = np.arange(num_seg * 2, dtype=int)
                src_idxs = get_src_indices_by_row(src_rows, shape=options['shape'])
                flat_src_idxs = True

                src = f'ode.{var}'
                tgt = f'path_constraints.all_values:{con_name}'

            phase.connect(src_name=src, tgt_name=tgt,
                          src_indices=src_idxs, flat_src_indices=flat_src_idxs)

    def setup_timeseries_outputs(self, phase):
        """
        Setup the timeseries for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        gd = self.grid_data

        for name, options in phase._timeseries.items():

            if options['transcription'] is None:
                ogd = None
            else:
                ogd = options['transcription'].grid_data

            timeseries_comp = RungeKuttaTimeseriesOutputComp(input_grid_data=gd,
                                                             output_grid_data=ogd,
                                                             output_subset=options['subset'])
            phase.add_subsystem(name, subsys=timeseries_comp)

    def configure_timeseries_outputs(self, phase):
        """
        Create connections from time series to all post-introspection sources.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        gd = self.grid_data
        num_seg = gd.num_segments
        time_units = phase.time_options['units']

        for timeseries_name, timeseries_options in phase._timeseries.items():
            timeseries_comp = phase._get_subsystem(timeseries_name)

            src_idxs = get_src_indices_by_row(gd.subset_node_indices['segment_ends'], (1,))

            timeseries_comp._add_output_configure('time',
                                                  shape=(1,),
                                                  units=time_units,
                                                  desc='time')

            timeseries_comp._add_output_configure('time_phase',
                                                  shape=(1,),
                                                  units=time_units,
                                                  desc='phase elapsed time')

            phase.connect(src_name='time', tgt_name=f'{timeseries_name}.input_values:time',
                          src_indices=src_idxs, flat_src_indices=True)

            phase.connect(src_name='time_phase', tgt_name=f'{timeseries_name}.input_values:time_phase',
                          src_indices=src_idxs, flat_src_indices=True)

            for state_name, options in phase.state_options.items():
                row_idxs = np.repeat(np.arange(1, num_seg, dtype=int), repeats=2)
                row_idxs = np.concatenate(([0], row_idxs, [num_seg]))
                src_idxs = get_src_indices_by_row(row_idxs, options['shape'])

                timeseries_comp._add_output_configure(f'states:{state_name}',
                                                      shape=options['shape'],
                                                      units=options['units'],
                                                      desc=options['desc'])

                phase.connect(src_name=f'states:{state_name}',
                              tgt_name=f'{timeseries_name}.input_values:states:{state_name}',
                              src_indices=src_idxs, flat_src_indices=True)

            for control_name, options in phase.control_options.items():
                control_units = options['units']
                src_rows = gd.subset_node_indices['segment_ends']
                src_idxs = get_src_indices_by_row(src_rows, options['shape'])

                timeseries_comp._add_output_configure(f'controls:{control_name}',
                                                      shape=options['shape'],
                                                      units=control_units,
                                                      desc=options['desc'])

                phase.connect(src_name='control_values:{0}'.format(control_name),
                              tgt_name=f'{timeseries_name}.input_values:controls:{control_name}',
                              src_indices=src_idxs, flat_src_indices=True)

                # Control rates
                timeseries_comp._add_output_configure(f'control_rates:{control_name}_rate',
                                                      shape=options['shape'],
                                                      units=get_rate_units(control_units,
                                                                           time_units,
                                                                           deriv=1),
                                                      desc=f'first time-derivative of control {control_name}')

                phase.connect(src_name=f'control_rates:{control_name}_rate',
                              tgt_name=f'{timeseries_name}.input_values:control_rates:{control_name}_rate',
                              src_indices=src_idxs, flat_src_indices=True)

                # Control second derivatives
                timeseries_comp._add_output_configure(f'control_rates:{control_name}_rate2',
                                                      shape=options['shape'],
                                                      units=get_rate_units(control_units,
                                                                           time_units,
                                                                           deriv=2),
                                                      desc=f'second time-derivative of control {control_name}')

                phase.connect(src_name=f'control_rates:{control_name}_rate2',
                              tgt_name=f'{timeseries_name}.input_values:control_rates:{control_name}_rate2',
                              src_indices=src_idxs, flat_src_indices=True)

            for control_name, options in phase.polynomial_control_options.items():
                src_rows = gd.subset_node_indices['segment_ends']
                src_idxs = get_src_indices_by_row(src_rows, options['shape'])

                control_units = options['units']

                timeseries_comp._add_output_configure(f'polynomial_controls:{control_name}',
                                                      shape=options['shape'],
                                                      units=control_units,
                                                      desc=options['desc'])

                phase.connect(src_name=f'polynomial_control_values:{control_name}',
                              tgt_name=f'{timeseries_name}.input_values:polynomial_controls:{control_name}',
                              src_indices=src_idxs, flat_src_indices=True)

                # Control rates
                timeseries_comp._add_output_configure('polynomial_control_rates:{0}_rate'.format(control_name),
                                                      shape=options['shape'],
                                                      units=get_rate_units(control_units, time_units,
                                                                           deriv=1),
                                                      desc=f'first time-derivative of polynomial '
                                                           f'control {control_name}')

                # Control rates
                phase.connect(src_name=f'polynomial_control_rates:{control_name}_rate',
                              tgt_name=f'{timeseries_name}.input_values:polynomial_control_rates:{control_name}_rate',
                              src_indices=src_idxs, flat_src_indices=True)

                # Control second derivatives
                timeseries_comp._add_output_configure(f'polynomial_control_rates:{control_name}_rate2',
                                                      shape=options['shape'],
                                                      units=get_rate_units(control_units, time_units,
                                                                           deriv=2),
                                                      desc=f'second time-derivative of polynomial '
                                                           f'control {control_name}')

                phase.connect(src_name=f'polynomial_control_rates:{control_name}_rate2',
                              tgt_name=f'{timeseries_name}.input_values:polynomial_control_rates:{control_name}_rate2',
                              src_indices=src_idxs, flat_src_indices=True)

            for param_name, options in phase.parameter_options.items():
                if options['include_timeseries']:
                    src_idxs_raw = np.zeros(self.grid_data.subset_num_nodes['segment_ends'], dtype=int)
                    src_idxs = get_src_indices_by_row(src_idxs_raw, options['shape'])

                    prom_name = 'parameters:{0}'.format(param_name)
                    tgt_name = 'input_values:parameters:{0}'.format(param_name)

                    timeseries_comp._add_output_configure(f'parameters:{param_name}',
                                                          shape=options['shape'],
                                                          units=options['units'],
                                                          desc='')

                    phase.promotes(timeseries_name, inputs=[(tgt_name, prom_name)],
                                   src_indices=src_idxs, flat_src_indices=True)

            for var, options in timeseries_options['outputs'].items():
                output_name = options['output_name']
                units = options.get('units', None)
                wildcard_units = options.get('wildcard_units', None)

                if '*' in var:  # match outputs from the ODE
                    ode_outputs = {opts['prom_name']: opts for (k, opts) in
                                   phase.ode.get_io_metadata(iotypes=('output',)).items()}
                    matches = filter(list(ode_outputs.keys()), var)
                else:
                    matches = [var]

                for v in matches:
                    if '*' in var:
                        output_name = v.split('.')[-1]
                        units = ode_outputs[v]['units']
                        # check for wildcard_units override of ODE units
                        if v in wildcard_units:
                            units = wildcard_units[v]

                    # Determine the path to the variable which we will be constraining
                    # This is more complicated for path constraints since, for instance,
                    # a single state variable has two sources which must be connected to
                    # the path component.
                    var_type = phase.classify_var(v)

                    # Ignore any variables that we've already added (states, times, controls, etc)
                    if var_type != 'ode':
                        continue

                    if self.is_static_ode_output(v, phase, 2*self.options['num_segments']):
                        warnings.warn(f'Cannot add ODE output {v} to the timeseries output. It is '
                                      f'sized such that its first dimension != num_nodes.')
                        continue

                    try:
                        shape, units = get_source_metadata(phase.ode, src=v,
                                                           user_units=units,
                                                           user_shape=options['shape'])
                    except ValueError:
                        raise ValueError(f'Timeseries output {v} is not a known variable in'
                                         f' the phase {phase.pathname} nor is it a known output of '
                                         f' the ODE.')

                    try:
                        timeseries_comp._add_output_configure(output_name, units, shape, desc='')
                    except ValueError as e:  # OK if it already exists
                        if 'already exists' in str(e):
                            continue
                        else:
                            raise e

                    # Failed to find variable, assume it is in the RHS
                    phase.connect(src_name=f'ode.{v}',
                                  tgt_name=f'{timeseries_name}.input_values:{output_name}')

    def get_parameter_connections(self, name, phase):
        """
        Returns info about a parameter's target connections in the phase.

        Parameters
        ----------
        name : str
            The name of the parameter for which connection info is desired.
        phase : dymos.Phase
            The Phase instance with which this transcription is associated.

        Returns
        -------
        list of (paths, indices)
            A list containing a tuple of target paths and corresponding src_indices to which the
            given design variable is to be connected.
        """
        connection_info = []
        num_seg = self.grid_data.num_segments
        num_stages = rk_methods[self.options['method']]['num_stages']
        num_iter_ode_nodes = num_seg * num_stages
        num_final_ode_nodes = 2 * num_seg

        if name in phase.parameter_options:
            options = phase.parameter_options[name]
            ode_tgts = get_targets(ode=phase.ode, name=name, user_targets=options['targets'])
            dynamic = options['dynamic']
            shape = options['shape']

            if dynamic:
                src_idxs_raw = np.zeros(num_final_ode_nodes, dtype=int)
                src_idxs = get_src_indices_by_row(src_idxs_raw, shape)
                if shape == (1,):
                    src_idxs = src_idxs.ravel()
            else:
                src_idxs_raw = np.zeros(1, dtype=int)
                src_idxs = get_src_indices_by_row(src_idxs_raw, shape)
                src_idxs = np.squeeze(src_idxs, axis=0)

            connection_info.append((['ode.{0}'.format(tgt) for tgt in ode_tgts], src_idxs))

            if dynamic:
                src_idxs_raw = np.zeros(num_iter_ode_nodes, dtype=int)
                src_idxs = get_src_indices_by_row(src_idxs_raw, shape)
                if shape == (1,):
                    src_idxs = src_idxs.ravel()
            else:
                src_idxs_raw = np.zeros(1, dtype=int)
                src_idxs = get_src_indices_by_row(src_idxs_raw, shape)
                src_idxs = np.squeeze(src_idxs, axis=0)

            connection_info.append((['rk_solve_group.ode.{0}'.format(tgt) for tgt in ode_tgts],
                                    src_idxs))

        return connection_info

    def _get_boundary_constraint_src(self, var, loc, phase):
        """
        Get the source for boundary constraint values.

        Parameters
        ----------
        var : str
            The name of the variable whose source information is desired.
        loc : str
            The location of the boundary constraint ['intitial', 'final'].
        phase : Phase
            The Phase instance with which this transcription is associated.

        Returns
        -------
        str
            Path to the source.
        shape
            Source shape.
        str
            Source units.
        bool
            True if the constraint is linear.
        """
        # Determine the path to the variable which we will be constraining
        time_units = phase.time_options['units']
        var_type = phase.classify_var(var)

        if var_type == 'time':
            shape = (1,)
            units = time_units
            linear = True
            constraint_path = 'time'
        elif var_type == 'time_phase':
            shape = (1,)
            units = time_units
            linear = True
            constraint_path = 'time_phase'
        elif var_type == 'state':
            shape = phase.state_options[var]['shape']
            units = phase.state_options[var]['units']
            linear = True if loc == 'initial' and phase.state_options[var]['fix_initial'] or \
                loc == 'final' and phase.state_options[var]['fix_final'] else False
            constraint_path = 'states:{0}'.format(var)
        elif var_type in ('indep_control', 'input_control'):
            shape = phase.control_options[var]['shape']
            units = phase.control_options[var]['units']
            linear = False
            constraint_path = 'control_values:{0}'.format(var)
        elif var_type in ('indep_polynomial_control', 'input_polynomial_control'):
            shape = phase.polynomial_control_options[var]['shape']
            units = phase.polynomial_control_options[var]['units']
            linear = False
            constraint_path = 'polynomial_control_values:{0}'.format(var)
        elif var_type == 'parameter':
            shape = phase.parameter_options[var]['shape']
            units = phase.parameter_options[var]['units']
            linear = True
            constraint_path = 'parameters:{0}'.format(var)
        elif var_type in ('control_rate', 'control_rate2'):
            control_var = var[:-5] if var_type == 'control_rate' else var[:-6]
            shape = phase.control_options[control_var]['shape']
            control_units = phase.control_options[control_var]['units']
            d = 1 if var_type == 'control_rate' else 2
            control_rate_units = get_rate_units(control_units, time_units, deriv=d)
            units = control_rate_units
            linear = False
            constraint_path = 'control_rates:{0}'.format(var)
        elif var_type in ('polynomial_control_rate', 'polynomial_control_rate2'):
            control_var = var[:-5] if var_type == 'polynomial_control_rate' else var[:-6]
            shape = phase.polynomial_control_options[control_var]['shape']
            control_units = phase.polynomial_control_options[control_var]['units']
            d = 1 if var_type == 'polynomial_control_rate' else 2
            control_rate_units = get_rate_units(control_units, time_units, deriv=d)
            units = control_rate_units
            linear = False
            constraint_path = 'polynomial_control_rates:{0}'.format(var)
        else:
            # Failed to find variable, assume it is in the ODE. This requires introspection.
            constraint_path = 'ode.{0}'.format(var)
            ode = phase._get_subsystem(self._rhs_source)
            shape, units = get_source_metadata(ode, src=var,
                                               user_shape=None,
                                               user_units=None)
            linear = False

        return constraint_path, shape, units, linear



import numpy as np
from scipy.linalg import block_diag

import openmdao.api as om

from ....utils.rk_methods import rk_methods
from ....options import options as dymos_options


class RungeKuttaStateAdvanceComp(om.ExplicitComponent):
    """
    Class definition for the RungeKuttaStateAdvanceComp.

    Given the initial value of each state at the start of each segment and the weight factors k
    for each state in each segment, compute the final value of each state at the end of each
    segment.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('num_segments', types=int,
                             desc='The number of segments (time steps) in the phase')

        self.options.declare('method', default='RK4', types=str,
                             desc='Specific Runge-Kutta Method to use.')

        self.options.declare('state_options', types=dict,
                             desc='Dictionary of state names/options for the phase')

    def configure_io(self):
        """
        I/O creation is delayed until configure so we can determine variable shape and units.
        """
        self._var_names = {}

        num_segs = self.options['num_segments']
        rk_data = rk_methods[self.options['method']]
        num_stages = rk_data['num_stages']

        for name, options in self.options['state_options'].items():
            shape = options['shape']
            size = np.prod(shape)
            units = options['units']

            self._var_names[name] = {}
            self._var_names[name]['initial'] = 'initial_states_per_seg:{0}'.format(name)
            self._var_names[name]['k'] = 'k:{0}'.format(name)
            self._var_names[name]['final'] = 'final_states:{0}'.format(name)
            self._var_names[name]['integral'] = 'state_integrals:{0}'.format(name)

            self.add_input(self._var_names[name]['initial'], shape=(num_segs,) + shape, units=units,
                           desc='The initial value of the state at the start of each segment.')

            self.add_input(self._var_names[name]['k'], shape=(num_segs, num_stages) + shape,
                           units=units, desc='RK multiplier k for each stage in each segment.')

            self.add_output(self._var_names[name]['integral'], shape=(num_segs,) + shape,
                            units=units,
                            desc='The change in value of the state along each segment')

            self.add_output(self._var_names[name]['final'], shape=(num_segs,) + shape,
                            units=units,
                            desc='The final value of the state at the end of each segment.')

            e = np.eye(size*num_segs)
            r, c = np.nonzero(e)
            self.declare_partials(of=self._var_names[name]['final'],
                                  wrt=self._var_names[name]['initial'],
                                  rows=r, cols=c, val=1.0)

            p = np.kron(rk_data['b'], np.eye(size))
            p = block_diag(*num_segs*[p])
            r, c = p.nonzero()

            self.declare_partials(of=self._var_names[name]['final'],
                                  wrt=self._var_names[name]['k'],
                                  rows=r, cols=c,
                                  val=np.tile(rk_data['b'], size*num_segs))

            self.declare_partials(of=self._var_names[name]['integral'],
                                  wrt=self._var_names[name]['k'],
                                  rows=r, cols=c,
                                  val=np.tile(rk_data['b'], size*num_segs))

    def compute(self, inputs, outputs):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        for name, options in self.options['state_options'].items():
            x0 = inputs[self._var_names[name]['initial']]
            k = inputs[self._var_names[name]['k']]
            b = rk_methods[self.options['method']]['b']
            final_name = self._var_names[name]['final']
            integral_name = self._var_names[name]['integral']
            outputs[integral_name] = np.einsum('ijk...,j...->ik...', k, b)
            outputs[final_name] = x0 + outputs[integral_name]



from .runge_kutta.runge_kutta import RungeKutta
from .pseudospectral.gauss_lobatto import GaussLobatto
from .pseudospectral.radau_pseudospectral import Radau
from .solve_ivp.solve_ivp import SolveIVP



from __future__ import print_function, division, absolute_import
import os
import unittest

import numpy as np
from numpy.testing import assert_almost_equal

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.examples.hyper_sensitive.hyper_sensitive_ode import HyperSensitiveODE
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE
from dymos.examples.brachistochrone.brachistochrone_vector_states_ode import BrachistochroneVectorStatesODE
from openmdao.utils.general_utils import set_pyoptsparse_opt
_, optimizer = set_pyoptsparse_opt('IPOPT', fallback=True)


@use_tempdirs
class TestRunProblem(unittest.TestCase):

    @unittest.skipIf(optimizer != 'IPOPT', 'IPOPT not available')
    def test_run_HS_problem_radau(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = optimizer

        if optimizer == 'SNOPT':
            p.driver.opt_settings['Major iterations limit'] = 200
            p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
            p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
        elif optimizer == 'IPOPT':
            p.driver.opt_settings['hessian_approximation'] = 'limited-memory'
            # p.driver.opt_settings['nlp_scaling_method'] = 'user-scaling'
            p.driver.opt_settings['print_level'] = 4
            p.driver.opt_settings['max_iter'] = 200
            p.driver.opt_settings['linear_solver'] = 'mumps'

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=HyperSensitiveODE,
                                                   transcription=dm.Radau(num_segments=10,
                                                                          order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=True)
        phase0.add_state('x', fix_initial=True, fix_final=False, rate_source='x_dot', targets=['x'])
        phase0.add_state('xL', fix_initial=True, fix_final=False, rate_source='L', targets=['xL'])
        phase0.add_control('u', opt=True, targets=['u'], rate_continuity=False)

        phase0.add_boundary_constraint('x', loc='final', equals=1)

        phase0.add_objective('xL', loc='final')

        phase0.set_refine_options(refine=True, tol=1e-6)

        p.setup(check=True)

        tf = np.float128(20)

        p.set_val('traj.phase0.states:x', phase0.interpolate(ys=[1.5, 1], nodes='state_input'))
        p.set_val('traj.phase0.states:xL', phase0.interpolate(ys=[0, 1], nodes='state_input'))
        p.set_val('traj.phase0.t_initial', 0)
        p.set_val('traj.phase0.t_duration', tf)
        p.set_val('traj.phase0.controls:u', phase0.interpolate(ys=[-0.6, 2.4],
                                                               nodes='control_input'))
        dm.run_problem(p, refine_method='hp', refine_iteration_limit=10)

        sqrt_two = np.sqrt(2)
        val = sqrt_two * tf
        c1 = (1.5 * np.exp(-val) - 1) / (np.exp(-val) - np.exp(val))
        c2 = (1 - 1.5 * np.exp(val)) / (np.exp(-val) - np.exp(val))

        ui = c1 * (1 + sqrt_two) + c2 * (1 - sqrt_two)
        uf = c1 * (1 + sqrt_two) * np.exp(val) + c2 * (1 - sqrt_two) * np.exp(-val)
        J = 0.5 * (c1 ** 2 * (1 + sqrt_two) * np.exp(2 * val) + c2 ** 2 * (1 - sqrt_two) *
                   np.exp(-2 * val) - (1 + sqrt_two) * c1 ** 2 - (1 - sqrt_two) * c2 ** 2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.controls:u')[0],
                          ui,
                          tolerance=5e-4)

        assert_near_equal(p.get_val('traj.phase0.timeseries.controls:u')[-1],
                          uf,
                          tolerance=5e-4)

        assert_near_equal(p.get_val('traj.phase0.timeseries.states:xL')[-1],
                          J,
                          tolerance=5e-4)

    @unittest.skipIf(optimizer != 'IPOPT', 'IPOPT not available')
    def test_run_HS_problem_gl(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = optimizer

        if optimizer == 'SNOPT':
            p.driver.opt_settings['Major iterations limit'] = 100
            p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
            p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
        elif optimizer == 'IPOPT':
            p.driver.opt_settings['hessian_approximation'] = 'limited-memory'
            # p.driver.opt_settings['nlp_scaling_method'] = 'user-scaling'
            p.driver.opt_settings['print_level'] = 4
            p.driver.opt_settings['linear_solver'] = 'mumps'

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=HyperSensitiveODE,
                                                   transcription=dm.GaussLobatto(num_segments=20,
                                                                                 order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=True)
        phase0.add_state('x', fix_initial=True, fix_final=False, rate_source='x_dot', targets=['x'])
        phase0.add_state('xL', fix_initial=True, fix_final=False, rate_source='L', targets=['xL'])
        phase0.add_control('u', opt=True, targets=['u'], rate_continuity=False)

        phase0.add_boundary_constraint('x', loc='final', equals=1)

        phase0.add_objective('xL', loc='final')

        phase0.set_refine_options(refine=True, tol=1.0E-5)

        p.setup(check=True)

        tf = 20

        p.set_val('traj.phase0.states:x', phase0.interpolate(ys=[1.5, 1], nodes='state_input'))
        p.set_val('traj.phase0.states:xL', phase0.interpolate(ys=[0, 1], nodes='state_input'))
        p.set_val('traj.phase0.t_initial', 0)
        p.set_val('traj.phase0.t_duration', tf)
        p.set_val('traj.phase0.controls:u', phase0.interpolate(ys=[-0.6, 2.4],
                                                               nodes='control_input'))
        dm.run_problem(p, refine_method='hp', refine_iteration_limit=5)

        sqrt_two = np.sqrt(2)
        val = sqrt_two * tf
        c1 = (1.5 * np.exp(-val) - 1) / (np.exp(-val) - np.exp(val))
        c2 = (1 - 1.5 * np.exp(val)) / (np.exp(-val) - np.exp(val))

        ui = c1 * (1 + sqrt_two) + c2 * (1 - sqrt_two)
        uf = c1 * (1 + sqrt_two) * np.exp(val) + c2 * (1 - sqrt_two) * np.exp(-val)
        J = 0.5 * (c1 ** 2 * (1 + sqrt_two) * np.exp(2 * val) + c2 ** 2 * (1 - sqrt_two) *
                   np.exp(-2 * val) - (1 + sqrt_two) * c1 ** 2 - (1 - sqrt_two) * c2 ** 2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.controls:u')[0],
                          ui,
                          tolerance=1e-2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.controls:u')[-1],
                          uf,
                          tolerance=1e-2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.states:xL')[-1],
                          J,
                          tolerance=5e-4)

    def test_run_brachistochrone_problem(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = 'SLSQP'

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=BrachistochroneODE,
                                                   transcription=dm.Radau(num_segments=10,
                                                                          order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=False)
        phase0.add_state('x', fix_initial=True, fix_final=False)
        phase0.add_state('y', fix_initial=True, fix_final=False)
        phase0.add_state('v', fix_initial=True, fix_final=False)
        phase0.add_control('theta', continuity=True, rate_continuity=True,
                           units='deg', lower=0.01, upper=179.9)
        phase0.add_parameter('g', units='m/s**2', val=9.80665)

        phase0.add_boundary_constraint('x', loc='final', equals=10)
        phase0.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase0.add_objective('time_phase', loc='final', scaler=10)

        phase0.set_refine_options(refine=True)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        p.set_val('traj.phase0.t_initial', 0.0)
        p.set_val('traj.phase0.t_duration', 2.0)

        p.set_val('traj.phase0.states:x', phase0.interpolate(ys=[0, 10], nodes='state_input'))
        p.set_val('traj.phase0.states:y', phase0.interpolate(ys=[10, 5], nodes='state_input'))
        p.set_val('traj.phase0.states:v', phase0.interpolate(ys=[0, 9.9], nodes='state_input'))
        p.set_val('traj.phase0.controls:theta', phase0.interpolate(ys=[5, 100],
                                                                   nodes='control_input'))
        p.set_val('traj.phase0.parameters:g', 9.80665)

        dm.run_problem(p)

        self.assertTrue(os.path.exists('dymos_solution.db'))
        # Assert the results are what we expect.
        cr = om.CaseReader('dymos_solution.db')
        case = cr.get_case('final')
        assert_almost_equal(case.outputs['traj.phase0.timeseries.time'].max(), 1.8016, decimal=4)

    def test_run_brachistochrone_vector_states_problem(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = 'SLSQP'

        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=dm.Radau(num_segments=1, order=3))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('pos', fix_initial=True, fix_final=[True, False])
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg', rate_continuity=False, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_boundary_constraint('pos', loc='final', equals=5, indices=[1])

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        pos0 = [0, 10]
        posf = [10, 5]

        p['phase0.states:pos'] = phase.interpolate(ys=[pos0, posf], nodes='state_input')
        p['phase0.states:v'] = phase.interpolate(ys=[0, 9.9], nodes='state_input')
        p['phase0.controls:theta'] = phase.interpolate(ys=[5, 100], nodes='control_input')
        p['phase0.parameters:g'] = 9.80665

        dm.run_problem(p, refine_iteration_limit=5)

        assert_near_equal(p.get_val('phase0.time')[-1], 1.8016, tolerance=1.0E-3)

    def test_run_brachistochrone_problem_with_simulate(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = 'SLSQP'

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=BrachistochroneODE,
                                                   transcription=dm.Radau(num_segments=10,
                                                                          order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=False)
        phase0.add_state('x', fix_initial=True, fix_final=False)
        phase0.add_state('y', fix_initial=True, fix_final=False)
        phase0.add_state('v', fix_initial=True, fix_final=False)
        phase0.add_control('theta', continuity=True, rate_continuity=True,
                           units='deg', lower=0.01, upper=179.9)
        phase0.add_parameter('g', units='m/s**2', val=9.80665)

        phase0.add_boundary_constraint('x', loc='final', equals=10)
        phase0.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase0.add_objective('time_phase', loc='final', scaler=10)

        phase0.set_refine_options(refine=True)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        p.set_val('traj.phase0.t_initial', 0.0)
        p.set_val('traj.phase0.t_duration', 2.0)

        p.set_val('traj.phase0.states:x', phase0.interpolate(ys=[0, 10], nodes='state_input'))
        p.set_val('traj.phase0.states:y', phase0.interpolate(ys=[10, 5], nodes='state_input'))
        p.set_val('traj.phase0.states:v', phase0.interpolate(ys=[0, 9.9], nodes='state_input'))
        p.set_val('traj.phase0.controls:theta', phase0.interpolate(ys=[5, 100],
                                                                   nodes='control_input'))
        p.set_val('traj.phase0.parameters:g', 9.80665)

        dm.run_problem(p, simulate=True)

        self.assertTrue(os.path.exists('dymos_solution.db'))
        # Assert the results are what we expect.
        cr = om.CaseReader('dymos_solution.db')
        case = cr.get_case('final')
        assert_almost_equal(case.outputs['traj.phase0.timeseries.time'].max(), 1.8016, decimal=4)

    def test_modify_problem(self):
        from dymos.examples.vanderpol.vanderpol_dymos import vanderpol
        from dymos.examples.vanderpol.vanderpol_dymos_plots import vanderpol_dymos_plots
        from dymos.run_problem import run_problem
        from scipy.interpolate import interp1d
        from numpy.testing import assert_almost_equal

        # Create the Dymos problem instance
        p = vanderpol(transcription='gauss-lobatto', num_segments=75)

        # Run the problem (simulate only)
        p.run_model()

        # simulate and record
        p.model.traj.simulate(record_file='vanderpol_simulation.sql')

        # create a new problem for restart to simulate a different command line execution
        q = vanderpol(transcription='gauss-lobatto', num_segments=75)

        # Call modify_problem with simulation restart database
        # modify_problem(q, restart='vanderpol_simulation.sql')

        # # Run the model
        run_problem(q, restart='vanderpol_simulation.sql')

        #  The solution should look like the explicit time history for the states and controls.
        DO_PLOTS = False
        if DO_PLOTS:
            vanderpol_dymos_plots(q)  # only for visual inspection and debug
        else:  # automate comparison
            s = q.model.traj.simulate()

            # get_val returns data for duplicate time points; remove them before interpolating
            tq = q.get_val('traj.phase0.timeseries.time')[:, 0]
            nodup = np.insert(tq[1:] != tq[:-1], 0, True)
            tq = tq[nodup]
            x1q = q.get_val('traj.phase0.timeseries.states:x1')[:, 0][nodup]
            x0q = q.get_val('traj.phase0.timeseries.states:x0')[:, 0][nodup]
            uq = q.get_val('traj.phase0.timeseries.controls:u')[:, 0][nodup]

            ts = s.get_val('traj.phase0.timeseries.time')[:, 0]
            nodup = np.insert(ts[1:] != ts[:-1], 0, True)
            ts = ts[nodup]
            x1s = s.get_val('traj.phase0.timeseries.states:x1')[:, 0][nodup]
            x0s = s.get_val('traj.phase0.timeseries.states:x0')[:, 0][nodup]
            us = s.get_val('traj.phase0.timeseries.controls:u')[:, 0][nodup]

            # create interpolation functions so that values can be looked up at matching time points
            fx1s = interp1d(ts, x1s, kind='cubic')
            fx0s = interp1d(ts, x0s, kind='cubic')
            fus = interp1d(ts, us, kind='cubic')

            assert_almost_equal(x1q, fx1s(tq), decimal=2)
            assert_almost_equal(x0q, fx0s(tq), decimal=2)
            assert_almost_equal(uq, fus(tq), decimal=5)


@use_tempdirs
class TestRunProblemPlotting(unittest.TestCase):
    def setUp(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = 'SLSQP'

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=BrachistochroneODE,
                                                   transcription=dm.Radau(num_segments=10,
                                                                          order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=True)
        phase0.add_state('x', fix_initial=True, fix_final=False)
        phase0.add_state('y', fix_initial=True, fix_final=False)
        phase0.add_state('v', fix_initial=True, fix_final=False)
        phase0.add_control('theta', continuity=True, rate_continuity=True,
                           units='deg', lower=0.01, upper=179.9)
        phase0.add_parameter('g', units='m/s**2', val=9.80665)

        phase0.add_boundary_constraint('x', loc='final', equals=10)
        phase0.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase0.add_objective('time_phase', loc='final', scaler=10)

        phase0.set_refine_options(refine=True)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        p.set_val('traj.phase0.t_initial', 0.0)
        p.set_val('traj.phase0.t_duration', 2.0)

        p.set_val('traj.phase0.states:x', phase0.interpolate(ys=[0, 10], nodes='state_input'))
        p.set_val('traj.phase0.states:y', phase0.interpolate(ys=[10, 5], nodes='state_input'))
        p.set_val('traj.phase0.states:v', phase0.interpolate(ys=[0, 9.9], nodes='state_input'))
        p.set_val('traj.phase0.controls:theta', phase0.interpolate(ys=[5, 100],
                                                                   nodes='control_input'))
        p.set_val('traj.phase0.parameters:g', 9.80665)

        self.p = p

    def test_run_brachistochrone_problem_make_plots(self):
        dm.run_problem(self.p, make_plots=True)

        for varname in ['time_phase', 'states:x', 'state_rates:x', 'states:y',
                        'state_rates:y', 'states:v',
                        'state_rates:v', 'controls:theta', 'control_rates:theta_rate',
                        'control_rates:theta_rate2', 'parameters:g']:
            self.assertTrue(os.path.exists(f'plots/{varname.replace(":","_")}.png'))

    def test_run_brachistochrone_problem_make_plots_set_plot_dir(self):
        plot_dir = "test_plot_dir"
        dm.run_problem(self.p, make_plots=True, plot_dir=plot_dir)

        for varname in ['time_phase', 'states:x', 'state_rates:x', 'states:y',
                        'state_rates:y', 'states:v',
                        'state_rates:v', 'controls:theta', 'control_rates:theta_rate',
                        'control_rates:theta_rate2', 'parameters:g']:
            self.assertTrue(os.path.exists(f'test_plot_dir/{varname.replace(":","_")}.png'))

    def test_run_brachistochrone_problem_do_not_make_plots(self):
        dm.run_problem(self.p, make_plots=False)

        for varname in ['time_phase', 'states:x', 'state_rates:x', 'states:y',
                        'state_rates:y', 'states:v',
                        'state_rates:v', 'controls:theta', 'control_rates:theta_rate',
                        'control_rates:theta_rate2', 'parameters:g']:
            self.assertFalse(os.path.exists(f'plots/{varname.replace(":","_")}.png'))

    def test_run_brachistochrone_problem_set_simulation_record_file(self):
        simulation_record_file = 'simulation_record_file.db'
        dm.run_problem(self.p, simulate=True, simulation_record_file=simulation_record_file)

        self.assertTrue(os.path.exists(simulation_record_file))

    def test_run_brachistochrone_problem_set_solution_record_file(self):
        solution_record_file = 'solution_record_file.db'
        dm.run_problem(self.p, solution_record_file=solution_record_file)

        self.assertTrue(os.path.exists(solution_record_file))

    def test_run_brachistochrone_problem_plot_simulation(self):
        dm.run_problem(self.p, make_plots=True, simulate=True)

        for varname in ['time_phase', 'states:x', 'state_rates:x', 'states:y',
                        'state_rates:y', 'states:v',
                        'state_rates:v', 'controls:theta', 'control_rates:theta_rate',
                        'control_rates:theta_rate2', 'parameters:g']:
            self.assertTrue(os.path.exists(f'plots/{varname.replace(":","_")}.png'))

    def test_run_brachistochrone_problem_plot_no_simulation_record_file_given(self):
        dm.run_problem(self.p, make_plots=True, simulate=True)

        for varname in ['time_phase', 'states:x', 'state_rates:x', 'states:y',
                        'state_rates:y', 'states:v',
                        'state_rates:v', 'controls:theta', 'control_rates:theta_rate',
                        'control_rates:theta_rate2', 'parameters:g']:
            self.assertTrue(os.path.exists(f'plots/{varname.replace(":","_")}.png'))


@use_tempdirs
class TestSimulateArrayParam(unittest.TestCase):

    def test_simulate_array_param(self):
        #
        # Initialize the Problem and the optimization driver
        #
        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        #
        # Create a trajectory and add a phase to it
        #
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = traj.add_phase('phase0',
                               dm.Phase(ode_class=BrachistochroneODE,
                                        transcription=dm.GaussLobatto(num_segments=10)))

        #
        # Set the variables
        #
        phase.set_time_options(initial_bounds=(0, 0), duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True, rate_source='xdot')

        phase.add_state('y', fix_initial=True, fix_final=True, rate_source='ydot')

        phase.add_state('v', fix_initial=True, fix_final=False, rate_source='vdot')

        phase.add_control('theta', continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665)
        phase.add_parameter('array', units=None, shape=(10,), dynamic=False)

        # dummy array of data
        indeps = p.model.add_subsystem('indeps', om.IndepVarComp(), promotes=['*'])
        indeps.add_output('array', np.linspace(1, 10, 10), units=None)
        # add dummy array as a parameter and connect it
        p.model.connect('array', 'traj.phase0.parameters:array')

        #
        # Minimize time at the end of the phase
        #
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        #
        # Setup the Problem
        #
        p.setup()

        #
        # Set the initial values
        #
        p['traj.phase0.t_initial'] = 0.0
        p['traj.phase0.t_duration'] = 2.0

        p['traj.phase0.states:x'] = phase.interpolate(ys=[0, 10], nodes='state_input')
        p['traj.phase0.states:y'] = phase.interpolate(ys=[10, 5], nodes='state_input')
        p['traj.phase0.states:v'] = phase.interpolate(ys=[0, 9.9], nodes='state_input')
        p['traj.phase0.controls:theta'] = phase.interpolate(ys=[5, 100.5],
                                                            nodes='control_input')

        #
        # Solve for the optimal trajectory
        #
        dm.run_problem(p, simulate=True)

        # Test the results
        sol_results = om.CaseReader('dymos_solution.db').get_case('final')
        sim_results = om.CaseReader('dymos_solution.db').get_case('final')

        sol = sol_results.get_val('traj.phase0.timeseries.parameters:array')
        sim = sim_results.get_val('traj.phase0.timeseries.parameters:array')

        assert_near_equal(sol - sim, np.zeros_like(sol))

        # Test that the parameter is available in the solution and simulation files
        sol = sol_results.get_val('traj.phase0.parameters:array')
        sim = sim_results.get_val('traj.phase0.parameters:array')

        assert_near_equal(sol - sim, np.zeros_like(sol))


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np

import openmdao.api as om
from ....utils.rk_methods import rk_methods
from ....utils.misc import get_rate_units
from ....options import options as dymos_options


class RungeKuttaKComp(om.ExplicitComponent):
    """
    Class definition for the RungeKuttaKComp.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('num_segments', types=int,
                             desc='The number of segments (timesteps) in the phase')

        self.options.declare('method', default='RK4', types=str,
                             desc='Specific Runge-Kutta Method to use.')

        self.options.declare('state_options', types=dict,
                             desc='Dictionary of state names/options for the phase')

        self.options.declare('time_units', default=None, allow_none=True, types=str,
                             desc='Units of the integration variable')

    def configure_io(self):
        """
        I/O creation is delayed until configure so we can determine variable shape and units.
        """
        self._var_names = {}

        num_seg = self.options['num_segments']
        rk_data = rk_methods[self.options['method']]
        num_stages = rk_data['num_stages']

        self.add_input('h', val=np.ones(num_seg), units=self.options['time_units'],
                       desc='step size for current Runge-Kutta segment.')

        for name, options in self.options['state_options'].items():
            shape = options['shape']
            units = options['units']
            rate_units = get_rate_units(units, self.options['time_units'])

            self._var_names[name] = {}
            self._var_names[name]['f'] = 'f:{0}'.format(name)
            self._var_names[name]['k'] = 'k:{0}'.format(name)

            self.add_input(self._var_names[name]['f'], shape=(num_seg, num_stages) + shape,
                           units=rate_units,
                           desc='The predicted values of the state at the ODE evaluation points.')

            self.add_output(self._var_names[name]['k'], shape=(num_seg, num_stages) + shape,
                            units=units, desc='RK multiplier k for each stage in the segment.')

            size = np.prod(shape)
            ar = np.arange(size * num_stages * num_seg, dtype=int)
            self.declare_partials(of=self._var_names[name]['k'],
                                  wrt=self._var_names[name]['f'],
                                  rows=ar, cols=ar)

            r = np.arange(size * num_stages * num_seg, dtype=int)
            c = np.repeat(np.arange(num_seg, dtype=int), num_stages * size)
            self.declare_partials(of=self._var_names[name]['k'],
                                  wrt='h',
                                  rows=r, cols=c)

    def compute(self, inputs, outputs):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        h = inputs['h']
        for name, options in self.options['state_options'].items():
            f = inputs[self._var_names[name]['f']]
            outputs[self._var_names[name]['k']] = f * h[:, np.newaxis, np.newaxis]

    def compute_partials(self, inputs, partials):
        """
        Compute sub-jacobian parts. The model is assumed to be in an unscaled state.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : Jacobian
            Subjac components written to partials[output_name, input_name].
        """
        num_stages = rk_methods[self.options['method']]['num_stages']
        h = inputs['h']
        for name, options in self.options['state_options'].items():
            size = np.prod(options['shape'])
            k_name = self._var_names[name]['k']
            f_name = self._var_names[name]['f']
            partials[k_name, f_name] = np.repeat(h, num_stages * size)
            partials[k_name, 'h'] = inputs[self._var_names[name]['f']].ravel()



import unittest

import numpy as np
from numpy.testing import assert_almost_equal

import openmdao.api as om
from dymos.utils.testing_utils import assert_check_partials

import dymos as dm
from dymos.transcriptions.common.boundary_constraint_comp import BoundaryConstraintComp

# Modify class so we can run it standalone.
from dymos.utils.misc import CompWrapperConfig
BoundaryConstraintComp = CompWrapperConfig(BoundaryConstraintComp)


class TestInitialScalarBoundaryValue(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True

        self.p = om.Problem(model=om.Group())

        ivp = self.p.model.add_subsystem('ivc', subsys=om.IndepVarComp(), promotes_outputs=['*'])

        ivp.add_output('x', val=np.arange(100))
        self.p.model.add_design_var('x', lower=0, upper=100)

        bv_comp = self.p.model.add_subsystem('bv_comp', BoundaryConstraintComp(loc='initial'))
        bv_comp._add_constraint(name='x', shape=(1,))

        self.p.model.connect('x', 'bv_comp.initial_value_in:x', src_indices=[0])

        self.p.setup(force_alloc_complex=True)
        self.p.run_model()

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_results(self):

        self.assertAlmostEqual(self.p['bv_comp.initial_value_in:x'][0], self.p['x'][0])
        self.assertAlmostEqual(self.p['bv_comp.initial_value:x'], self.p['x'][0])

    def test_partials(self):
        cpd = self.p.check_partials(compact_print=True, method='cs')
        assert_check_partials(cpd)


class TestFinalScalarBoundaryValue(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True

        self.p = om.Problem(model=om.Group())

        ivp = self.p.model.add_subsystem('ivc', subsys=om.IndepVarComp(), promotes_outputs=['*'])

        ivp.add_output('x', val=np.arange(100))
        self.p.model.add_design_var('x', lower=0, upper=100)

        bv_comp = self.p.model.add_subsystem('bv_comp', BoundaryConstraintComp(loc='final'))
        bv_comp._add_constraint(name='x', shape=(1,))

        self.p.model.connect('x', 'bv_comp.final_value_in:x', src_indices=[-1])

        self.p.setup(force_alloc_complex=True)
        self.p.run_model()

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_results(self):

        self.assertAlmostEqual(self.p['bv_comp.final_value_in:x'][0], self.p['x'][-1])
        self.assertAlmostEqual(self.p['bv_comp.final_value:x'], self.p['x'][-1])

    def test_partials(self):
        cpd = self.p.check_partials(compact_print=True, method='cs')
        assert_check_partials(cpd)


class TestVectorInitialBoundaryValue(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True

        self.p = om.Problem(model=om.Group())

        ivp = self.p.model.add_subsystem('ivc', subsys=om.IndepVarComp(), promotes_outputs=['*'])

        ivp.add_output('pos', val=np.zeros((100, 3)))
        self.p.model.add_design_var('pos', lower=0, upper=100)

        bv_comp = self.p.model.add_subsystem('bv_comp', BoundaryConstraintComp(loc='initial'))
        bv_comp._add_constraint(name='pos', shape=(3,))

        src_idxs = np.array([0, 1, 2], dtype=int)

        self.p.model.connect('pos', 'bv_comp.initial_value_in:pos', src_indices=src_idxs,
                             flat_src_indices=True)

        self.p.setup(force_alloc_complex=True)

        self.p['pos'][:, 0] = np.arange(100)
        self.p['pos'][:, 1] = 100 + np.arange(100)
        self.p['pos'][:, 2] = 200 + np.arange(100)

        self.p.run_model()

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_results(self):
        assert_almost_equal(self.p['bv_comp.initial_value_in:pos'], self.p['pos'][0, :])
        assert_almost_equal(self.p['bv_comp.initial_value:pos'], self.p['pos'][0, :])

    def test_partials(self):
        cpd = self.p.check_partials(compact_print=True, method='cs')
        assert_check_partials(cpd)


class TestVectorFinalBoundaryValue(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True

        self.p = om.Problem(model=om.Group())

        ivp = self.p.model.add_subsystem('ivc', subsys=om.IndepVarComp(), promotes_outputs=['*'])

        ivp.add_output('pos', val=np.zeros((100, 3)))
        self.p.model.add_design_var('pos', lower=0, upper=100)

        bv_comp = self.p.model.add_subsystem('bv_comp', BoundaryConstraintComp(loc='final'))
        bv_comp._add_constraint(name='pos', shape=(3,))

        src_idxs = np.array([-3, -2, -1])

        self.p.model.connect('pos', 'bv_comp.final_value_in:pos', src_indices=src_idxs,
                             flat_src_indices=True)

        self.p.setup(force_alloc_complex=True)

        self.p['pos'][:, 0] = np.arange(100)
        self.p['pos'][:, 1] = 100 + np.arange(100)
        self.p['pos'][:, 2] = 200 + np.arange(100)

        self.p.run_model()

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_results(self):

        assert_almost_equal(self.p['bv_comp.final_value_in:pos'], self.p['pos'][-1, :])
        assert_almost_equal(self.p['bv_comp.final_value:pos'], self.p['pos'][-1, :])

    def test_partials(self):
        cpd = self.p.check_partials(compact_print=True, method='cs')
        assert_check_partials(cpd)


class TestMatrixInitialBoundaryValue(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True

        self.p = om.Problem(model=om.Group())

        ivp = self.p.model.add_subsystem('ivc', subsys=om.IndepVarComp(), promotes_outputs=['*'])

        ivp.add_output('M', val=np.zeros((100, 3, 3)))
        self.p.model.add_design_var('M', lower=0, upper=100)

        bv_comp = self.p.model.add_subsystem('bv_comp', BoundaryConstraintComp(loc='initial'))
        bv_comp._add_constraint(name='M', shape=(3, 3))

        src_idxs = np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])

        self.p.model.connect('M', 'bv_comp.initial_value_in:M', src_indices=src_idxs,
                             flat_src_indices=True)

        self.p.setup(force_alloc_complex=True)

        self.p['M'][:, 0, 0] = np.arange(100)
        self.p['M'][:, 0, 1] = 100 + np.arange(100)
        self.p['M'][:, 0, 2] = 200 + np.arange(100)

        self.p['M'][:, 1, 0] = 1000 + np.arange(100)
        self.p['M'][:, 1, 1] = 1000 + 100 + np.arange(100)
        self.p['M'][:, 1, 2] = 1000 + 200 + np.arange(100)

        self.p['M'][:, 2, 0] = 2000 + np.arange(100)
        self.p['M'][:, 2, 1] = 2000 + 100 + np.arange(100)
        self.p['M'][:, 2, 2] = 2000 + 200 + np.arange(100)

        self.p.run_model()

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_results(self):

        assert_almost_equal(self.p['bv_comp.initial_value_in:M'], self.p['M'][0, ...])
        assert_almost_equal(self.p['bv_comp.initial_value:M'], self.p['M'][0, ...])

    def test_partials(self):
        cpd = self.p.check_partials(compact_print=True, method='cs')
        assert_check_partials(cpd)


class TestMatrixFinalBoundaryValue(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True

        self.p = om.Problem(model=om.Group())

        ivp = self.p.model.add_subsystem('ivc', subsys=om.IndepVarComp(), promotes_outputs=['*'])

        ivp.add_output('M', val=np.zeros((100, 3, 3)))
        self.p.model.add_design_var('M', lower=0, upper=100)

        bv_comp = self.p.model.add_subsystem('bv_comp', BoundaryConstraintComp(loc='final'))
        bv_comp._add_constraint(name='M', shape=(3, 3))

        src_idxs = np.array([[-9, -8, -7], [-6, -5, -4], [-3, -2, -1]])

        self.p.model.connect('M', 'bv_comp.final_value_in:M', src_indices=src_idxs,
                             flat_src_indices=True)

        self.p.setup(force_alloc_complex=True)

        self.p['M'][:, 0, 0] = np.arange(100)
        self.p['M'][:, 0, 1] = 100 + np.arange(100)
        self.p['M'][:, 0, 2] = 200 + np.arange(100)

        self.p['M'][:, 1, 0] = 1000 + np.arange(100)
        self.p['M'][:, 1, 1] = 1000 + 100 + np.arange(100)
        self.p['M'][:, 1, 2] = 1000 + 200 + np.arange(100)

        self.p['M'][:, 2, 0] = 2000 + np.arange(100)
        self.p['M'][:, 2, 1] = 2000 + 100 + np.arange(100)
        self.p['M'][:, 2, 2] = 2000 + 200 + np.arange(100)

        self.p.run_model()

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_results(self):

        assert_almost_equal(self.p['bv_comp.final_value_in:M'], self.p['M'][-1, ...])
        assert_almost_equal(self.p['bv_comp.final_value:M'], self.p['M'][-1, ...])

    def test_partials(self):
        cpd = self.p.check_partials(compact_print=True, method='cs')
        assert_check_partials(cpd)


class TestMultipleConstraints(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True

        self.p = om.Problem(model=om.Group())

        ivp = self.p.model.add_subsystem('ivc', subsys=om.IndepVarComp(), promotes_outputs=['*'])

        ivp.add_output('M', val=np.zeros((100, 3, 3)))
        ivp.add_output('pos', val=np.zeros((100, 3)))
        ivp.add_output('x', val=np.zeros((100,)))

        self.p.model.add_design_var('M', lower=0, upper=100)
        self.p.model.add_design_var('pos', lower=0, upper=100)
        self.p.model.add_design_var('x', lower=0, upper=100)

        bv_comp = self.p.model.add_subsystem('bv_comp', BoundaryConstraintComp(loc='initial'))

        M_shape = (3, 3)
        M_size = np.prod(M_shape)
        bv_comp._add_constraint(name='M', shape=M_shape)
        src_idxs = np.arange(M_size, dtype=int).reshape(M_shape)
        self.p.model.connect('M', 'bv_comp.initial_value_in:M', src_indices=src_idxs,
                             flat_src_indices=True)

        pos_shape = (3,)
        pos_size = np.prod(pos_shape)
        bv_comp._add_constraint(name='pos', shape=pos_shape)
        src_idxs = np.arange(pos_size, dtype=int).reshape(pos_shape)
        self.p.model.connect('pos', 'bv_comp.initial_value_in:pos', src_indices=src_idxs,
                             flat_src_indices=True)

        x_shape = (1,)
        x_size = np.prod(x_shape)
        bv_comp._add_constraint(name='x', shape=x_shape)
        src_idxs = np.arange(x_size, dtype=int).reshape(x_shape)
        self.p.model.connect('x', 'bv_comp.initial_value_in:x', src_indices=src_idxs,
                             flat_src_indices=True)

        self.p.setup(force_alloc_complex=True)

        self.p['M'][:, 0, 0] = np.arange(100)
        self.p['M'][:, 0, 1] = 100 + np.arange(100)
        self.p['M'][:, 0, 2] = 200 + np.arange(100)

        self.p['M'][:, 1, 0] = 1000 + np.arange(100)
        self.p['M'][:, 1, 1] = 1000 + 100 + np.arange(100)
        self.p['M'][:, 1, 2] = 1000 + 200 + np.arange(100)

        self.p['M'][:, 2, 0] = 2000 + np.arange(100)
        self.p['M'][:, 2, 1] = 2000 + 100 + np.arange(100)
        self.p['M'][:, 2, 2] = 2000 + 200 + np.arange(100)

        self.p['pos'][:, 0] = 100000 + np.arange(100)
        self.p['pos'][:, 1] = 100000 + 100 + np.arange(100)
        self.p['pos'][:, 2] = 100000 + 200 + np.arange(100)

        self.p.run_model()

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_results(self):

        assert_almost_equal(self.p['bv_comp.initial_value_in:M'], self.p['M'][0, ...])
        assert_almost_equal(self.p['bv_comp.initial_value:M'], self.p['M'][0, ...])

        assert_almost_equal(self.p['bv_comp.initial_value_in:pos'], self.p['pos'][0, ...])
        assert_almost_equal(self.p['bv_comp.initial_value:pos'], self.p['pos'][0, ...])

        assert_almost_equal(self.p['bv_comp.initial_value_in:x'], self.p['x'][0, ...])
        assert_almost_equal(self.p['bv_comp.initial_value:x'], self.p['x'][0, ...])

    def test_partials(self):
        cpd = self.p.check_partials(compact_print=True, method='cs')
        assert_check_partials(cpd)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

from scipy.interpolate import interp1d

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


@use_tempdirs
class TestTimeseriesOutput(unittest.TestCase):

    def test_timeseries_gl(self, test_smaller_timeseries=False):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=8, order=3, compressed=True))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=True,
                          units='deg', lower=0.01, upper=179.9, ref=1, ref0=0)

        if test_smaller_timeseries:
            phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665, include_timeseries=False)
        else:
            phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        p['phase0.states:x'] = phase.interpolate(ys=[0, 10], nodes='state_input')
        p['phase0.states:y'] = phase.interpolate(ys=[10, 5], nodes='state_input')
        p['phase0.states:v'] = phase.interpolate(ys=[0, 9.9], nodes='state_input')
        p['phase0.controls:theta'] = phase.interpolate(ys=[5, 100], nodes='control_input')
        p['phase0.parameters:g'] = 9.80665

        p.run_driver()

        gd = phase.options['transcription'].grid_data
        state_input_idxs = gd.subset_node_indices['state_input']
        control_input_idxs = gd.subset_node_indices['control_input']
        col_idxs = gd.subset_node_indices['col']

        assert_near_equal(p.get_val('phase0.time'),
                          p.get_val('phase0.timeseries.time')[:, 0])

        assert_near_equal(p.get_val('phase0.time_phase'),
                          p.get_val('phase0.timeseries.time_phase')[:, 0])

        for state in ('x', 'y', 'v'):
            assert_near_equal(p.get_val('phase0.states:{0}'.format(state)),
                              p.get_val('phase0.timeseries.states:'
                                        '{0}'.format(state))[state_input_idxs])

            assert_near_equal(p.get_val('phase0.state_interp.state_col:{0}'.format(state)),
                              p.get_val('phase0.timeseries.states:'
                                        '{0}'.format(state))[col_idxs])

        for control in ('theta',):
            assert_near_equal(p.get_val('phase0.controls:{0}'.format(control)),
                              p.get_val('phase0.timeseries.controls:'
                                        '{0}'.format(control))[control_input_idxs])

        for dp in ('g',):
            for i in range(gd.subset_num_nodes['all']):
                if test_smaller_timeseries:
                    with self.assertRaises(KeyError):
                        p.get_val('phase0.timeseries.parameters:{0}'.format(dp))
                else:
                    assert_near_equal(p.get_val('phase0.parameters:{0}'.format(dp))[0],
                                      p.get_val('phase0.timeseries.parameters:{0}'.format(dp))[i])

        # call simulate to test SolveIVP transcription
        exp_out = phase.simulate()
        if test_smaller_timeseries:
            with self.assertRaises(KeyError):
                exp_out.get_val('phase0.timeseries.parameters:{0}'.format(dp))
        else:  # no error accessing timseries.parameter
            exp_out.get_val('phase0.timeseries.parameters:{0}'.format(dp))

    def test_timeseries_gl_smaller_timeseries(self):
        self.test_timeseries_gl(test_smaller_timeseries=True)

    def test_timeseries_radau(self, test_smaller_timeseries=False):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Radau(num_segments=8, order=3, compressed=True))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=True,
                          units='deg', lower=0.01, upper=179.9, ref=1, ref0=0)

        if test_smaller_timeseries:
            phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665, include_timeseries=False)
        else:
            phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.model.options['assembled_jac_type'] = 'csc'
        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        p['phase0.states:x'] = phase.interpolate(ys=[0, 10], nodes='state_input')
        p['phase0.states:y'] = phase.interpolate(ys=[10, 5], nodes='state_input')
        p['phase0.states:v'] = phase.interpolate(ys=[0, 9.9], nodes='state_input')
        p['phase0.controls:theta'] = phase.interpolate(ys=[5, 100], nodes='control_input')
        p['phase0.parameters:g'] = 9.80665

        p.run_driver()

        gd = phase.options['transcription'].grid_data
        state_input_idxs = gd.subset_node_indices['state_input']
        control_input_idxs = gd.subset_node_indices['control_input']

        assert_near_equal(p.get_val('phase0.time'),
                          p.get_val('phase0.timeseries.time')[:, 0])

        assert_near_equal(p.get_val('phase0.time_phase'),
                          p.get_val('phase0.timeseries.time_phase')[:, 0])

        for state in ('x', 'y', 'v'):
            assert_near_equal(p.get_val('phase0.states:{0}'.format(state)),
                              p.get_val('phase0.timeseries.states:'
                                        '{0}'.format(state))[state_input_idxs])

        for control in ('theta',):
            assert_near_equal(p.get_val('phase0.controls:{0}'.format(control)),
                              p.get_val('phase0.timeseries.controls:'
                                        '{0}'.format(control))[control_input_idxs])

        for dp in ('g',):
            for i in range(gd.subset_num_nodes['all']):
                if test_smaller_timeseries:
                    with self.assertRaises(KeyError):
                        p.get_val('phase0.timeseries.parameters:{0}'.format(dp))
                else:
                    assert_near_equal(p.get_val('phase0.parameters:{0}'.format(dp))[0],
                                      p.get_val('phase0.timeseries.parameters:'
                                                '{0}'.format(dp))[i])

        # call simulate to test SolveIVP transcription
        exp_out = phase.simulate()
        if test_smaller_timeseries:
            with self.assertRaises(KeyError):
                exp_out.get_val('phase0.timeseries.parameters:{0}'.format(dp))
        else:  # no error accessing timseries.parameter
            exp_out.get_val('phase0.timeseries.parameters:{0}'.format(dp))

        # Test that the state rates are output in both the radau and solveivp timeseries outputs
        t_sol = p.get_val('phase0.timeseries.time')
        t_sim = exp_out.get_val('phase0.timeseries.time')

        for state_name in ('x', 'y', 'v'):
            rate_sol = p.get_val(f'phase0.timeseries.state_rates:{state_name}')
            rate_sim = exp_out.get_val(f'phase0.timeseries.state_rates:{state_name}')

            rate_t_sim = interp1d(t_sim.ravel(), rate_sim.ravel())

            assert_near_equal(rate_t_sim(t_sol), rate_sol, tolerance=1.0E-3)

    def test_timeseries_radau_smaller_timeseries(self):
        self.test_timeseries_radau(test_smaller_timeseries=True)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import openmdao.api as om
from openmdao.utils.general_utils import set_pyoptsparse_opt
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


@use_tempdirs
class TestPhaseParameterPromotion(unittest.TestCase):

    def test_promotes_parameter(self):
        transcription = 'radau-ps'
        optimizer = 'SNOPT'
        num_segments = 10
        transcription_order = 3
        compressed = False

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        OPT, OPTIMIZER = set_pyoptsparse_opt(optimizer, fallback=True)
        p.driver.options['optimizer'] = OPTIMIZER
        p.driver.declare_coloring()

        if transcription == 'gauss-lobatto':
            t = dm.GaussLobatto(num_segments=num_segments,
                                order=transcription_order,
                                compressed=compressed)
        elif transcription == 'radau-ps':
            t = dm.Radau(num_segments=num_segments,
                         order=transcription_order,
                         compressed=compressed)
        elif transcription == 'runge-kutta':
            t = dm.RungeKutta(num_segments=num_segments,
                              order=transcription_order,
                              compressed=compressed)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=t)

        traj.add_phase('phase0', phase, promotes_inputs=['t_initial', 't_duration', 'parameters:g'])

        p.model.add_subsystem('traj', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='s')

        phase.add_state('x', fix_initial=True, fix_final=False, solve_segments=False,
                        units='m', rate_source='xdot')
        phase.add_state('y', fix_initial=True, fix_final=False, solve_segments=False,
                        units='m', rate_source='ydot')
        phase.add_state('v', fix_initial=True, fix_final=False, solve_segments=False,
                        units='m/s', rate_source='vdot', targets=['v'])

        phase.add_control('theta', continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9, targets=['theta'])

        phase.add_parameter('g', units='m/s**2', val=9.80665, targets=['g'])

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        p['traj.t_initial'] = 0.0
        p['traj.t_duration'] = 2.0

        p['traj.phase0.states:x'] = phase.interpolate(ys=[0, 10], nodes='state_input')
        p['traj.phase0.states:y'] = phase.interpolate(ys=[10, 5], nodes='state_input')
        p['traj.phase0.states:v'] = phase.interpolate(ys=[0, 9.9], nodes='state_input')
        p['traj.phase0.controls:theta'] = phase.interpolate(ys=[5, 100], nodes='control_input')
        p['traj.parameters:g'] = 9.80665

        p.run_driver()

        assert_near_equal(p['traj.t_duration'], 1.8016, tolerance=1.0E-4)


if __name__ == '__main__':
    unittest.main()
